#if defined (CONFIG_UNIT_TEST)
#include "windows.h"
#include "tchar.h"
#include "immdev.h"
#include "..\skki1_0.h"
#include "..\resource.h"

static	int		srnMyColorIndex2SysColorIndexTbl []	= {
	COLOR_BTNFACE,
	COLOR_BTNTEXT,
	COLOR_ACTIVEBORDER,
	COLOR_ACTIVECAPTION,
	COLOR_CAPTIONTEXT,
	COLOR_APPWORKSPACE,
	COLOR_WINDOW,
	COLOR_WINDOWTEXT,
	COLOR_DESKTOP,
	COLOR_INFOBK,
	COLOR_INFOTEXT,
	COLOR_WINDOWTEXT,
	COLOR_MENU,
	COLOR_MENUTEXT,
	COLOR_HIGHLIGHTTEXT,
	COLOR_HIGHLIGHT,
	COLOR_INACTIVEBORDER,
	COLOR_INACTIVECAPTION,
	COLOR_INACTIVECAPTIONTEXT,
} ;

COLORREF
GetImeColor (
	register HWND	hwnd,
	register int	nColor)
{
	register HDC		hdc ;
	register COLORREF	col ;

	switch (nColor) {
	case	MYCOLOR_TEXTAUTO:
	case	MYCOLOR_BACKAUTO:
		hdc	= GetDC (hwnd) ;
		col	= GetBkColor (hdc) ;
		ReleaseDC (hwnd, hdc) ;
		return	(nColor == MYCOLOR_BACKAUTO)? col : ~col ;
	case	MYCOLOR_BLACK:
		return	PALETTERGB (  0,   0,   0) ;
	case	MYCOLOR_DARKRED:
		return	PALETTERGB (128,   0,   0) ;
	case	MYCOLOR_DARKGREEN:
		return	PALETTERGB (  0, 128,   0) ;
	case	MYCOLOR_DARKYELLOW:
		return	PALETTERGB (128, 128,   0) ;
	case	MYCOLOR_DARKBLUE:
		return	PALETTERGB (  0,   0, 128) ;
	case	MYCOLOR_DARKPURPLE:
		return	PALETTERGB (128,   0, 128) ;
	case	MYCOLOR_DARKLIGHTBLUE:
		return	PALETTERGB (0,   128, 128) ;
	case	MYCOLOR_DARKGRAY:
		return	PALETTERGB (128, 128, 128) ;
	case	MYCOLOR_LIGHTGRAY:
		return	PALETTERGB (192, 192, 192) ;
	case	MYCOLOR_RED:
		return	PALETTERGB (255,   0,   0) ;
	case	MYCOLOR_GREEN:
		return	PALETTERGB (  0, 255,   0) ;
	case	MYCOLOR_YELLOW:
		return	PALETTERGB (255, 255,   0) ;
	case	MYCOLOR_BLUE:
		return	PALETTERGB (  0,   0, 255) ;
	case	MYCOLOR_PURPLE:
		return	PALETTERGB (255,   0, 255) ;
	case	MYCOLOR_LIGHTBLUE:
		return	PALETTERGB (  0, 255, 255) ;
	case	MYCOLOR_WHITE:
		return	PALETTERGB (255, 255, 255) ;
	default:
		if (MYCOLOR_SYSTEM <= nColor && nColor < MAX_MYCOLOR) {
			nColor	-= MYCOLOR_SYSTEM ;
			return	GetSysColor (srnMyColorIndex2SysColorIndexTbl [nColor]) ;
		}
		break ;
	}
	return	PALETTERGB (0, 0, 0) ;
}

/*	$BE@@~$NIA2h$K;H$&%Q%?!<%s!#(B*/
static const UINT		srbDottedBrush [8]	= {
	0xAAAAAAAA, 0xAAAAAAAA, 0xAAAAAAAA, 0xAAAAAAAA,
	0xAAAAAAAA, 0xAAAAAAAA, 0xAAAAAAAA, 0xAAAAAAAA,
} ;

/*	$B%G%#%6@~$NIA2h$K;H$&%Q%?!<%s!#(B*/
static const UINT		srbDitherBrush [8]	= {
	0xAAAA5555, 0x5555AAAA, 0xAAAA5555, 0x5555AAAA,
	0xAAAA5555, 0x5555AAAA, 0xAAAA5555, 0x5555AAAA,
} ;

HBRUSH
GetImeLineBrush (
	register HWND				hwnd,		/* [in] */
	register int				nLineColor,	/* [in] */
	register int				nLineType,	/* [in] */
	register COLORREF*			pColBrush,	/* [out] */
	register HBITMAP*			phBm,		/* [out] */
	register int*				pnWidth)	/* [out] */
{
	register COLORREF	colBrush ;
	register HBRUSH		hBrush	= NULL ;
	register HBITMAP	hBitmap	= NULL ;

	if (pColBrush == NULL || phBm == NULL || pnWidth == NULL) {
		return	NULL ;
	}

	colBrush	= GetImeColor (hwnd, nLineColor) ;
	switch (nLineType) {
	case	MYLINE_SOLID:
		*pnWidth	= 1 ;
		hBrush		= CreateSolidBrush (colBrush) ;
		break ;
	case	MYLINE_DOTTED:
		hBitmap		= CreateBitmap (8, 8, 1, 1, (LPBYTE)srbDottedBrush) ;
		hBrush		= CreatePatternBrush (hBitmap) ;
		*pnWidth	= 1 ;
		break ;
	case	MYLINE_THICK_SOLID:
		*pnWidth	= 2 ;
		hBrush		= CreateSolidBrush (colBrush) ;
		break ;
	case	MYLINE_THIN_DITHER:
		*pnWidth	= 2 ;
		hBitmap		= CreateBitmap (8, 8, 1, 1, (LPBYTE)srbDitherBrush) ;
		hBrush		= CreatePatternBrush (hBitmap) ;
		break ;
	case	MYLINE_THICK_DITHER:
		*pnWidth	= 3 ;
		hBitmap		= CreateBitmap (8, 8, 1, 1, (LPBYTE)srbDitherBrush) ;
		hBrush		= CreatePatternBrush (hBitmap) ;
		break ;
	case	MYLINE_NO:
	default:
		*pnWidth	= 0 ;
		break ;
	}
	*pColBrush	= colBrush ;
	*phBm		= hBitmap ;
	return	hBrush ;
}

long	PASCAL
Myatol (LPMYSTR lpString)
{
	long	lValue ;
	long	lSign ;

	lValue	= 0 ;
	if (!lpString || !*lpString)
		return	lValue ;
	if (*lpString == MYTEXT ('+')){
		lSign	= 1 ;
		lpString	++ ;
	} else if (*lpString == MYTEXT ('-')){
		lSign	= -1 ;
		lpString	++ ;
	} else {
		lSign	= 1 ;
	}
	while (*lpString && MYTEXT ('0') <= *lpString && *lpString <= MYTEXT ('9')){
		lValue	= lValue * 10 + (*lpString - MYTEXT ('0')) ;
		lpString	++ ;
	}
	return	lSign * lValue ;
}

#if defined (UNICODE)
int	PASCAL	MylstrcmpW (
	register LPCWSTR	lp0,
	register LPCWSTR	lp1)
{
	while(*lp0 && *lp1 && (*lp0 == *lp1)) {
		lp0++;
		lp1++;
	}
	return (*lp0 - *lp1);
}

int PASCAL
MylstrncmpW (
	register LPCWSTR	lp0,
	register LPCWSTR	lp1, 
	register int		iNum)
{
	while(iNum > 0 && *lp0 && *lp1 && (*lp0 == *lp1)){
		lp0 ++ ;
		lp1 ++ ;
		iNum -- ;
	}
	return (iNum > 0)? (*lp0 - *lp1) : 0 ;
}

int	PASCAL
MylstrcpyW (LPWSTR lp0, LPCWSTR lp1)
{
	int n	= 0 ;

	while(*lp1){
		*lp0	= *lp1 ;
		lp0 ++ ;
		lp1 ++ ;
		n ++ ;
	}
	*lp0	= *lp1 ;
	return	n ;
}

int	PASCAL
MylstrncpyW (
	register LPWSTR		lp0,
	register LPCWSTR	lp1,
	register int		iNum)
{
	int	iN	= iNum ;
	while (iNum > 0 && *lp1){
		*lp0 ++	= *lp1 ++ ;
		iNum	-- ;
	}
	if (iNum > 0){
		*lp0	= *lp1 ;
	}
	return	(iN - iNum) ;
}

int	PASCAL
MylstrcatW (LPWSTR lp0, LPCWSTR lp1)
{
	int n	= 0 ;
	while (*lp0)
		lp0 ++ ;
	while(*lp1){
		*lp0	= *lp1 ;
		lp0 ++ ;
		lp1 ++ ;
		n ++ ;
	}
	*lp0	= *lp1 ;
	return	n ;
}

int	PASCAL
MylstrncatW (LPWSTR lp0, LPCWSTR lp1, int iNum)
{
	int	iN	= iNum ;
	while (*lp0)
		lp0	++ ;
	while (iNum > 0 && *lp1){
		*lp0 ++	= *lp1 ++ ;
		iNum	-- ;
	}
	if (iNum > 0){
		*lp0	= *lp1 ;
	}
	return	(iN - iNum) ;
}

LPWSTR PASCAL
MyCharPrevW(LPWSTR lpStart, LPWSTR lpCur)
{
	LPWSTR lpRet = lpStart;
	if (lpCur > lpStart)
		lpRet = lpCur - 1;

	return lpRet;
}

LPWSTR
PASCAL MyCharNextW(LPWSTR lp)
{
	return lp++;
}

LPWSTR
PASCAL MylstrcpynW(LPWSTR lp0, LPWSTR lp1, int nCount)
{
	int	n ;
	for (n = 0 ; *lp1 && n < nCount - 1; *lp0++ = *lp1++, n++)
		;
	*lp0 = L'\0' ;
	return	lp0 ;
}

#else	/* defined (UNICODE) */

int PASCAL MylstrncmpA (LPMYSTR lp0, LPMYSTR lp1, int iNum)
{
	while(iNum > 0 && *lp0 && *lp1 && (*lp0 == *lp1)){
		lp0 ++ ;
		lp1 ++ ;
		iNum -- ;
	}
	return (iNum > 0)? (*lp0 - *lp1) : 0 ;
}

int	PASCAL	MylstrncpyA (LPMYSTR lp0, LPMYSTR lp1, int iNum)
{
	int	iN	= iNum ;
	while (iNum > 0 && *lp1){
		*lp0 ++	= *lp1 ++ ;
		iNum	-- ;
	}
	if (iNum > 0){
		*lp0	= *lp1 ;
	}
	return	(iN - iNum) ;
}

int	PASCAL	MylstrncatA (LPMYSTR lp0, LPMYSTR lp1, int iNum)
{
	int	iN	= iNum ;
	while (*lp0)
		lp0	++ ;
	while (iNum > 0 && *lp1){
		*lp0 ++	= *lp1 ++ ;
		iNum	-- ;
	}
	if (iNum > 0){
		*lp0	= *lp1 ;
	}
	return	(iN - iNum) ;
}

#endif

void	PASCAL
Mysnprintf (
	LPTSTR		lpBuffer,
	int			nBuffer,
	LPCTSTR		lpFormat,
	...)
{
	register LPCTSTR	pSrc ;
	register LPTSTR		pDest ;
	register int		nDest ;
	TCHAR				szBuffer [128] ;
	va_list				vl ;

	if (nBuffer <= 0 || lpBuffer == NULL)
		return ;

	va_start (vl, lpFormat) ;

	pSrc	= lpFormat ;
	pDest	= lpBuffer ;
	nDest	= nBuffer - 1 ;
	while (nDest > 0 && *pSrc != TEXT ('\0')) {
		switch (*pSrc) {
			register LPCTSTR	pString ;
			register int		nString, nCopy ;

			/*	% $B$N=hM}!#(B*/
		case	TEXT ('%'):
			pSrc	++ ;
			switch (*pSrc) {
			case	TEXT ('s'):
			{
				pString	= va_arg (vl, LPCTSTR) ;
				goto	common_copy ;
			}
			case	TEXT ('u'):
			{
				register unsigned int	uNumber	= va_arg (vl, unsigned int) ;

				/*	wsprintf $B$O(B CRT $B4X?t$G$O$"$k$,!"(B
				 *	mainCRTStartup $B$+$i(B _main $B$r;2>H$9$k$h$&$JL5Cc$O(B
				 *	$B$7$J$$!#(B*/
				wsprintf (szBuffer, TEXT ("%u"), uNumber) ;
				pString	= szBuffer ;
				goto	common_copy ;
			}
			case	TEXT ('d'):
			{
				register int		nNumber	= va_arg (vl, int) ;

				wsprintf (szBuffer, TEXT ("%d"), nNumber) ;
				pString	= szBuffer ;
				goto	common_copy ;
			}
			common_copy:
			nString	= lstrlen (pString) ;
			nCopy	= (nDest < nString)? nDest : nString ;
			memcpy (pDest, pString, nCopy * sizeof (TCHAR)) ;
			pDest	+= nCopy ;
			nDest	-= nCopy ;
			pSrc	++ ;
			break ;

			case	TEXT ('%'):
			{
				*pDest	++	= TEXT ('%') ;
				nDest	-- ;
				pSrc	++ ;
				break ;
			}
			default:
				if (*pSrc != TEXT ('\0'))
					pSrc	++ ;
				break ;
			}
			break ;
			
		default:
			*pDest ++	= *pSrc ++ ;
			nDest -- ;
			break ;
		}
	}
	va_end (vl) ;

	*pDest	= TEXT ('\0') ;
	return ;
}

#endif

