/* # SKKIME (Simple Kana-Kanji Input Method Editor for Windows'98)
 * Config.c
 * This file is part of SKKIME.
 * Copyright (C) 1999
 * Takashi SAKAMOTO (tatari_sakamoto@nifty.ne.jp)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with skkinput; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "windows.h"
#include "tchar.h"
#include "immdev.h"
#include "commctrl.h"
#include "skki1_0.h"
#include <limits.h>
#include <stddef.h>
#include "conf\skimcnfp.h"
#include "resource.h"

#define	REGPATH_COLORFACE	MYTEXT("Software\\TT\\Windows\\CurrentVersion\\SKKIME\\1.0\\ColorFace")
#define	REGSUBKEY_COLORFACE	MYTEXT("style")

typedef struct tagMYCOLORDEF {
	LPMYSTR		m_strText ;
	int			m_nType ;
}	MYCOLORDEF ;

/*========================================================================*
 *	prototypes
 */
static	INT_PTR	skkimeConfig_dlgColorFaceOnInitDialog	(HWND, UINT, WPARAM, LPARAM) ;
static	INT_PTR	skkimeConfig_dlgColorFaceOnNotify		(HWND, UINT, WPARAM, LPARAM) ;
static	INT_PTR	skkimeConfig_dlgColorFaceOnCommand		(HWND, UINT, WPARAM, LPARAM) ;

static	BOOL	skkimeConfig_setupColorComboBox		(HWND) ;
static	BOOL	skkimeConfig_setupLineTypeComboBox	(HWND) ;
static	void	skkimeConfig_setColorComboBox		(HWND, const MYCOLORFACESET*) ;

static	BOOL	skkimeConfig_initColorFace (TINYMEMMGR*, LPCMYSTR, MYCOLORFACESET*, const MYCOLORFACESET*, int) ;
static	BOOL	skkimeConfig_updateColorFace (LPCMYSTR, MYCOLORFACESET*) ;
static	BOOL	skkimeConfig_isValidColor	(int) ;
static	BOOL	skkimeConfig_isValidLineType	(int) ;

/*========================================================================*
 *	global variables
 */
static	MYCOLORFACESET		srSkkImeDefaultColorFace []	= {
	/*	ϊB*/
	{ MYCOLOR_TEXTAUTO,			/* () */	/* FB*/
	  MYCOLOR_BACKAUTO,			/* (wi) */ 	/* wiFB*/
	  MYCOLOR_TEXTAUTO,			/* () */	/* FB*/
	  MYLINE_THIN_DITHER,		/* fBU() */	/* ^CvB*/
	},
	/*	ϊB*/
	{ MYCOLOR_HIGHLIGHTTEXT,	/* I() */	/* FB*/
	  MYCOLOR_HIGHLIGHT,		/* I(wi) */	/* wiFB*/
	  MYCOLOR_TEXTAUTO,			/* () */	 	/* FB*/
	  MYLINE_NO,				/* Ȃ */				/* ^CvB*/
	},
} ;

static	MYCOLORDEF		srSkkImeColorDefinition []	= {
	{ MYTEXT ("()"),					MYCOLOR_TEXTAUTO, },
	{ MYTEXT ("(wi)"),					MYCOLOR_BACKAUTO, },
	{ MYTEXT (""),							MYCOLOR_BLACK, },
	{ MYTEXT ("Â"),						MYCOLOR_DARKRED, },	
	{ MYTEXT ("Â"),						MYCOLOR_DARKGREEN, },
	{ MYTEXT ("ÂF"),						MYCOLOR_DARKYELLOW, },
	{ MYTEXT ("Â"),						MYCOLOR_DARKBLUE, },
	{ MYTEXT ("Â"),						MYCOLOR_DARKPURPLE, },
	{ MYTEXT ("ÂF"),						MYCOLOR_DARKLIGHTBLUE, },
	{ MYTEXT ("ÂDF"),						MYCOLOR_DARKGRAY, },
	{ MYTEXT ("邢DF"),					MYCOLOR_LIGHTGRAY, },
	{ MYTEXT (""),							MYCOLOR_RED, },
	{ MYTEXT (""),							MYCOLOR_GREEN, },
	{ MYTEXT ("F"),							MYCOLOR_YELLOW, },
	{ MYTEXT (""),							MYCOLOR_BLUE, },
	{ MYTEXT (""),							MYCOLOR_PURPLE, },
	{ MYTEXT ("F"),							MYCOLOR_LIGHTBLUE, },
	{ MYTEXT (""),							MYCOLOR_WHITE, },
	{ MYTEXT ("3DIuWFNg"),				MYCOLOR_BTNFACE, },
	{ MYTEXT ("3DIuWFNg̕"),			MYCOLOR_BTNTEXT, },
	{ MYTEXT ("ANeBu EBhE̋E"),	MYCOLOR_ACTIVEBORDER, },
	{ MYTEXT ("ANeBu ^Cg o["),		MYCOLOR_ACTIVECAPTION, },
	{ MYTEXT ("ANeBu ^Cg o[̕"),MYCOLOR_CAPTIONTEXT, },
	{ MYTEXT ("AvP[Vƈ"),		MYCOLOR_APPWORKSPACE, },
	{ MYTEXT ("EBhE"),					MYCOLOR_WINDOW, },
	{ MYTEXT ("EBhE̕"),				MYCOLOR_WINDOWTEXT, },
	{ MYTEXT ("fXNgbv"),					MYCOLOR_DESKTOP, },
	{ MYTEXT ("qg"),						MYCOLOR_INFOBK, },
	{ MYTEXT ("qg̕"),					MYCOLOR_INFOTEXT, },
	{ MYTEXT ("bZ[W{bNX̕"),		MYCOLOR_MSGBOXTEXT, },
	{ MYTEXT ("j["),						MYCOLOR_MENU, },
	{ MYTEXT ("j[̕"),				MYCOLOR_MENUTEXT, },
	{ MYTEXT ("I()"),				MYCOLOR_HIGHLIGHTTEXT, },
	{ MYTEXT ("I(wi)"),				MYCOLOR_HIGHLIGHT, },
	{ MYTEXT ("ANeBu EBhE̋E"),	MYCOLOR_INACTIVEBORDER, },
	{ MYTEXT ("ANeBu ^Cg o["),	MYCOLOR_INACTIVECAPTION, },
	{ MYTEXT ("ANeBu ^Cg o[̕"),	MYCOLOR_INACTIVECAPTIONTEXT, },
} ;

BOOL
skkimeConfig_InitializeColorFace (
	register TINYMEMMGR*		pMemMgr,
	register MYCOLORFACESET*	pColorFaceSet)
{
	if (pMemMgr == NULL || pColorFaceSet == NULL)
		return	FALSE ;

	return	skkimeConfig_initColorFace (pMemMgr, REGPATH_COLORFACE, pColorFaceSet, srSkkImeDefaultColorFace, ARRAYSIZE (srSkkImeDefaultColorFace)) ;
}

BOOL
skkimeConfig_UpdateColorFace (
	register MYCOLORFACESET*		pColorFace)
{
	if (pColorFace == NULL)
		return	FALSE ;
	return	skkimeConfig_updateColorFace (REGPATH_COLORFACE, pColorFace) ;
}

INT_PTR	CALLBACK
skkimeConfig_dlgColorFaceTblProc (
	register HWND		hDlg,
	register UINT		uMsg,
	register WPARAM		wParam,
	register LPARAM		lParam)
{
	switch (uMsg) {
	case	WM_INITDIALOG:
		return	skkimeConfig_dlgColorFaceOnInitDialog (hDlg, uMsg, wParam, lParam) ;
	case	WM_NOTIFY:
		return	skkimeConfig_dlgColorFaceOnNotify (hDlg, uMsg, wParam, lParam) ;
	case	WM_COMMAND:
		return	skkimeConfig_dlgColorFaceOnCommand (hDlg, uMsg, wParam, lParam) ;
	default:
		break ;
	}
	return	FALSE ;
}

/*========================================================================
 *	private functions
 */
INT_PTR
skkimeConfig_dlgColorFaceOnInitDialog (
	register HWND		hDlg,
	register UINT		uMsg,
	register WPARAM		wParam,
	register LPARAM		lParam)
{
	register LPPROPSHEETPAGE	pPropSheet ;
	register SkkimeConfigArg*	pConfArg ;
	register int	n ;
	MYLVITEM		lvI ;
	static TMYLVCOLUMNDEF		srColumnDef []	= {
		{ 120,	MYTEXT ("ύX"), },
		{ 60,	MYTEXT ("vr["), },	/*	preview  owner-draw ̕Kv肩H */
	} ;
	static LPMYSTR				srColorFaceTitle []	= {
		MYTEXT ("ϊ"),	MYTEXT ("ϊ"),
	} ;
	register HWND	hwndLV, hwndCB ;
	register MYCOLORFACESET*	pColor ;

	DEBUGPRINTF ((MYTEXT ("skkimeConfig_dlgColorFaceOnInitDialog (%x)\n"), hDlg)) ;

	SetWindowLongPtr (hDlg, DWLP_USER, (LONG_PTR)lParam) ;
	pPropSheet	= (LPPROPSHEETPAGE) lParam ;
	pConfArg	= (SkkimeConfigArg *) pPropSheet->lParam ;

	hwndLV	= GetDlgItem (hDlg, IDC_LIST_SELECTFACE) ;
	if (hwndLV != NULL) {
		MYLV_COLUMN		lvColumn ;
		register int	i, nItem ;
		POINT			pt ;

		MyListView_DeleteAllItems (hwndLV) ;
		lvColumn.mask		= LVCF_FMT | LVCF_WIDTH | LVCF_TEXT | LVCF_SUBITEM ;
		lvColumn.fmt		= LVCFMT_LEFT ;
		for (i = 0 ; i < ARRAYSIZE (srColumnDef) ; i ++) {
			lvColumn.cx			= srColumnDef [i].m_nWidth ;
			lvColumn.pszText	= (LPMYSTR) srColumnDef [i].m_strText ;
			MyListView_InsertColumn (hwndLV, i, &lvColumn) ;
		}

		memset (&lvI, 0, sizeof (lvI)) ;
		lvI.mask		= LVIF_PARAM | LVIF_STATE ;
		n				= 0 ;
		for (i = 0 ; i < MAX_SKKIME_COLORFACE ; i ++) {
			lvI.iItem		= i ;
			lvI.lParam		= (LPARAM) i ;
			nItem			= MyListView_InsertItem (hwndLV, &lvI) ;
			if (nItem >= 0) {
				MyListView_SetItemText (hwndLV, nItem, 0, srColorFaceTitle [i]) ;
				MyListView_SetItemText (hwndLV, nItem, 1, MYTEXT ("")) ;
			}
		}
		MyListView_SetSelectionMark (hwndLV, nItem) ;
		MyListView_SetItemState (hwndLV, nItem, LVIS_FOCUSED | LVIS_SELECTED, 0) ;
		MyListView_GetItemPosition (hwndLV, 0, &pt) ;
		MySendMessage (hwndLV, WM_LBUTTONDOWN, 0, (LPARAM)(((DWORD)(WORD)pt.y << 16) | (DWORD)(WORD)pt.x)) ;
	}
	pColor			= &pConfArg->m_rColorFaceSet [0] ;
	hwndCB			= GetDlgItem (hDlg, IDC_COMBO_FORECOLOR) ;
	skkimeConfig_setupColorComboBox (hwndCB) ;
	MySendMessage (hwndCB, CB_SETCURSEL, (WPARAM) pColor->m_nTextColor, (LPARAM) 0) ;
	hwndCB			= GetDlgItem (hDlg, IDC_COMBO_BACKCOLOR) ;
	skkimeConfig_setupColorComboBox (hwndCB) ;
	MySendMessage (hwndCB, CB_SETCURSEL, (WPARAM) pColor->m_nBackColor, (LPARAM) 0) ;
	hwndCB			= GetDlgItem (hDlg, IDC_COMBO_ULCOLOR) ;
	skkimeConfig_setupColorComboBox (hwndCB) ;
	MySendMessage (hwndCB, CB_SETCURSEL, (WPARAM) pColor->m_nUnderLineColor, (LPARAM) 0) ;
	hwndCB			= GetDlgItem (hDlg, IDC_COMBO_UNDERLINE) ;
	skkimeConfig_setupLineTypeComboBox (hwndCB) ;
	MySendMessage (hwndCB, CB_SETCURSEL, (WPARAM) pColor->m_nUnderLineType, (LPARAM) 0) ;
	return	TRUE ;

	UNREFERENCED_PARAMETER (uMsg) ;
	UNREFERENCED_PARAMETER (wParam) ;
}

INT_PTR
skkimeConfig_dlgColorFaceOnNotify (
	register HWND		hDlg,
	register UINT		uMsg,
	register WPARAM		wParam,
	register LPARAM		lParam)
{
	register NMHDR	FAR*		pNMHDR	= (NMHDR *) lParam ;
	register LPPROPSHEETPAGE	pPropSheet ;
	register SkkimeConfigArg*	pConfArg ;
	register MYCOLORFACESET*	pColor ;

	pPropSheet	= (LPPROPSHEETPAGE) GetWindowLongPtr (hDlg, DWLP_USER) ;
	pConfArg	= (SkkimeConfigArg *) pPropSheet->lParam ;

	switch (pNMHDR->idFrom) {
	case	IDC_COMBO_FORECOLOR:
	case	IDC_COMBO_BACKCOLOR:
	case	IDC_COMBO_ULCOLOR:
	case	IDC_COMBO_UNDERLINE:
	{
		/*	 preview ̏ԂύXȂ΂ȂȂB*/
		return	FALSE ;
	}
	case	IDC_LIST_SELECTFACE:
	{
		switch (pNMHDR->code) {
			/*	ListView  customdraw ̂߂̏B
			 */
		case	NM_CUSTOMDRAW:
		{
			register LPNMLVCUSTOMDRAW	lpLVCD	= (LPNMLVCUSTOMDRAW) lParam ;
			register LRESULT			lRetval	= 0 ;

			switch (lpLVCD->nmcd.dwDrawStage) {
			case	CDDS_PREPAINT:
				lRetval		= CDRF_NOTIFYITEMDRAW ;
				break ;
			case	CDDS_ITEMPREPAINT:
				lRetval		= CDRF_NOTIFYSUBITEMDRAW ;
				break ;
			case	CDDS_ITEMPREPAINT | CDDS_SUBITEM:
			{
			    register int		nItem		= lpLVCD->nmcd.dwItemSpec ;
				register int		nSubItem	= lpLVCD->iSubItem ;
				static   LPCTSTR	str			= TEXT ("") ;
				register HWND		hwnd		= pNMHDR->hwndFrom ;
				register HDC		hDC			= lpLVCD->nmcd.hdc ;

				if (nItem < 0 || nItem >= MAX_SKKIME_COLORFACE)
					break ;

				pColor		= &pConfArg->m_rColorFaceSet [nItem] ;
				if (nSubItem == 1) {
					RECT		rc ;
					SIZE		sz ;
					HBRUSH		hBrush ;
					HBITMAP		hBitmap ;
					int			nWidth, nBkMode ;
					COLORREF	colLine, colTextBak, colBkBak ;

					/*	Tvuv`悷B
					 */
					colTextBak	= SetTextColor (hDC, TSkkIme_crGetImeColor (hwnd, pColor->m_nTextColor)) ;
					colBkBak	= SetBkColor   (hDC, TSkkIme_crGetImeColor (hwnd, pColor->m_nBackColor)) ;
					nBkMode		= SetBkMode	 (hDC, OPAQUE) ;
					ListView_GetItemRect (hwnd, nItem, &rc, LVIR_BOUNDS) ;
					rc.left		= lpLVCD->nmcd.rc.left ;
					rc.right	= lpLVCD->nmcd.rc.right ;
					DrawText (hDC, str, lstrlen (str), &rc, DT_LEFT | DT_TOP) ;
					
					/*	TvuvɉB
					 */
					GetTextExtentPoint32 (hDC, str, lstrlen (str), &sz) ;
					hBrush	= TSkkIme_hbrGetImeLineBrush (hwnd, pColor->m_nUnderLineColor, pColor->m_nUnderLineType, &colLine, &hBitmap, &nWidth) ; 
					if (hBrush != NULL) {
						register HBRUSH	hOldBrush ;
						register int		nCX, nX, nY ;

						nX	= rc.left ;
						nCX	= ((nX + sz.cx) > rc.right)? (rc.right - nX) : sz.cx ;
						nY	= ((rc.top + sz.cy) > rc.bottom)? rc.bottom : (rc.top + sz.cy) ;
						nY	-= nWidth ;
						hOldBrush		= (HBRUSH) SelectObject (hDC, hBrush) ;
						SetTextColor (hDC, colLine) ;
						PatBlt (hDC, nX, nY, nCX, nWidth, PATCOPY) ;
						(void) SelectObject (hDC, hOldBrush) ;
						DeleteObject (hBrush) ;
					}
					if (hBitmap != NULL) 
						DeleteObject (hBitmap) ;
					lRetval		= CDRF_SKIPDEFAULT ;
					SetTextColor (hDC, colTextBak) ;
					SetBkColor (hDC, colBkBak) ;
					SetBkMode	 (hDC, nBkMode) ;
				} else {
					lRetval		= CDRF_NOTIFYSUBITEMDRAW ;
				}
				break ;
			}
			default:
				lRetval		= CDRF_DODEFAULT ;
				break ;
			}
			SetWindowLongPtr (hDlg, DWLP_MSGRESULT, lRetval) ;
			return	TRUE ;
		}

		/*	ListView őIĂ item ύXꂽꍇɁA
		 *	E ComboBox ̕\XVȂ΂ȂȂB
		 */
		case	NM_CLICK:
		case	NM_DBLCLK:
		case	NM_RCLICK:
		case	NM_RDBLCLK:
		{
			register LPNMITEMACTIVATE	lpnmitem = (LPNMITEMACTIVATE) lParam ;
			
			DEBUGPRINTF ((TEXT ("NM_CLICK (item:%d)\n"), lpnmitem->iItem)) ;

			if (lpnmitem->iItem < 0 || 
				lpnmitem->iItem >= MAX_SKKIME_COLORFACE)
				break ;
			pColor	= &pConfArg->m_rColorFaceSet [lpnmitem->iItem] ;
			skkimeConfig_setColorComboBox (hDlg, pColor) ;
			return	0 ;
		}
		default:
			break ;
		}
		break ;
	}
	default:
		switch (pNMHDR->code){
		case PSN_SETACTIVE:
			SetWindowLongPtr (hDlg, DWLP_MSGRESULT, 0) ;
			return	TRUE ;

		case PSN_KILLACTIVE:
			SetWindowLongPtr (hDlg, DWLP_MSGRESULT, FALSE) ;
			return	TRUE ;

		case PSN_RESET:
		case PSN_HELP:
			break ;

		case PSN_APPLY:
			if (skkimeConfig_UpdateColorFace (pConfArg->m_rColorFaceSet)) {
				SetWindowLongPtr (hDlg, DWLP_MSGRESULT, PSNRET_NOERROR) ;
			} else {
				SetWindowLongPtr (hDlg, DWLP_MSGRESULT, PSNRET_INVALID) ;
			}
			return	TRUE ;

		default:
			return	FALSE ;
		}
		return	TRUE ;
	}
	return	TRUE ;

	UNREFERENCED_PARAMETER (hDlg) ;
	UNREFERENCED_PARAMETER (uMsg) ;
	UNREFERENCED_PARAMETER (wParam) ;
}

INT_PTR
skkimeConfig_dlgColorFaceOnCommand (
	register HWND		hDlg,
	register UINT		uMsg,
	register WPARAM		wParam,
	register LPARAM		lParam)
{
	register LPPROPSHEETPAGE	pPropSheet ;
	register SkkimeConfigArg*	pConfArg ;
	register WORD	wNotifyCode ;
	register WORD	wID ;
	register HWND	hwndCtl ;
	//register HWND	hwndLV ;

	pPropSheet	= (LPPROPSHEETPAGE) GetWindowLongPtr (hDlg, DWLP_USER) ;
	pConfArg	= (SkkimeConfigArg *) pPropSheet->lParam ;

	wNotifyCode	= HIWORD (wParam) ;
	wID			= LOWORD (wParam) ;
	hwndCtl		= (HWND) lParam ;

	if (wNotifyCode == CBN_SELCHANGE) {
		register HWND	hwndLV	= GetDlgItem (hDlg, IDC_LIST_SELECTFACE) ;
		register MYCOLORFACESET*	pColor ;
		register int	nItem ;
		register long	lResult ;

		if (hwndLV == NULL)
			return	0 ;
		nItem	= MyListView_GetSelectionMark (hwndLV) ;
		if (nItem < 0 || nItem >= MAX_SKKIME_COLORFACE)
			return	0 ;
		pColor	= &pConfArg->m_rColorFaceSet [nItem] ;
		lResult	= SendMessage (hwndCtl, CB_GETCURSEL, 0, 0) ;
		switch (wID) {
		case	IDC_COMBO_FORECOLOR:
			pColor->m_nTextColor	= lResult ;
			break ;
		case	IDC_COMBO_BACKCOLOR:
			pColor->m_nBackColor	= lResult ;
			break ;
		case	IDC_COMBO_ULCOLOR:
			pColor->m_nUnderLineColor	= lResult ;
			break ;
		case	IDC_COMBO_UNDERLINE:
			pColor->m_nUnderLineType	= lResult ;
			break ;
		default:
			return	0 ;
		}
		/*	`AXVB*/
		PropSheet_Changed (pConfArg->m_hPropSheetDlg, hDlg) ;
		InvalidateRect (hwndLV, NULL, FALSE) ;
		return	1 ;
	}
	return	0 ;
	UNREFERENCED_PARAMETER (hDlg) ;
	UNREFERENCED_PARAMETER (uMsg) ;
}

BOOL
skkimeConfig_setupColorComboBox (
	register HWND		hwndCB)
{
	register int			i, nRetval ;
	register MYCOLORDEF*	ptr ;

	if (hwndCB == NULL || ! IsWindow (hwndCB))
		return	FALSE ;

	ptr	= srSkkImeColorDefinition ;
	for (i = 0 ; i < ARRAYSIZE (srSkkImeColorDefinition) ; i ++) {
		nRetval	= MySendMessage (hwndCB, CB_INSERTSTRING, i, (LPARAM) ptr->m_strText) ;
		ptr	++ ;
	}
	return	TRUE ;
}

BOOL
skkimeConfig_setupLineTypeComboBox (
	register HWND		hwndCB)
{
	static LPCMYSTR		srString []	= {
		MYTEXT ("Ȃ"),	MYTEXT (""),		MYTEXT ("_"),
		MYTEXT (""),	MYTEXT ("fBUא"),	MYTEXT ("fBU"),
	} ;
	register int		i ;

	if (hwndCB == NULL || ! IsWindow (hwndCB))
		return	FALSE ;

	for (i = 0 ; i < ARRAYSIZE (srString) ; i ++) 
		MySendMessage (hwndCB, CB_INSERTSTRING, i, (LPARAM) srString [i]) ;
	return	TRUE ;
}

void
skkimeConfig_setColorComboBox (
	register HWND					hDlg,
	register const MYCOLORFACESET*	pColor)
{
	register HWND	hwndCB ;

	if (hDlg == NULL || pColor == NULL)
		return ;

	hwndCB	= GetDlgItem (hDlg, IDC_COMBO_FORECOLOR) ;
	MySendMessage (hwndCB, CB_SETCURSEL, (WPARAM) pColor->m_nTextColor, 0) ;
	hwndCB	= GetDlgItem (hDlg, IDC_COMBO_BACKCOLOR) ;
	MySendMessage (hwndCB, CB_SETCURSEL, (WPARAM) pColor->m_nBackColor, 0) ;
	hwndCB	= GetDlgItem (hDlg, IDC_COMBO_ULCOLOR) ;
	MySendMessage (hwndCB, CB_SETCURSEL, (WPARAM) pColor->m_nUnderLineColor, 0) ;
	hwndCB	= GetDlgItem (hDlg, IDC_COMBO_UNDERLINE) ;
	MySendMessage (hwndCB, CB_SETCURSEL, (WPARAM) pColor->m_nUnderLineType, 0) ;
	return ;
}

BOOL
skkimeConfig_initColorFace (
	register TINYMEMMGR*			pMemMgr,
	register LPCMYSTR				strRegPath,
	register MYCOLORFACESET*		pColorFaceSet,
	register const MYCOLORFACESET*	pColorFaceSetDefault,
	register int					nColorFaceSetDefault)
{
	HKEY			hSubKey ;
	BYTE			rbyStyle [MAX_SKKIME_COLORFACE * 4] ;
	register BOOL	fInitialized	= FALSE ;

	if (pMemMgr == NULL ||
		strRegPath == NULL ||
		pColorFaceSetDefault == NULL ||
		nColorFaceSetDefault <= 0)
		return	FALSE ;

	if (MyRegOpenKeyEx (HKEY_CURRENT_USER, strRegPath, 0, KEY_READ, &hSubKey) == ERROR_SUCCESS) {
		DWORD	cbData, dwType ;
		
		DEBUGPRINTF ((MYTEXT ("RegKeyOpenEx succeeded (\"%s\")\n"), strRegPath)) ;

		/*	j-map ̏B*/
		cbData	= sizeof (rbyStyle) ;
		if (MyRegQueryValueEx (hSubKey, REGSUBKEY_COLORFACE, NULL, &dwType, rbyStyle, &cbData) == ERROR_SUCCESS &&
			dwType == REG_BINARY &&
			cbData == sizeof (rbyStyle)) {
			register const MYCOLORFACESET*	pDefault	= pColorFaceSetDefault ;
			register MYCOLORFACESET*		pColor		= pColorFaceSet ;
			register LPBYTE					pByte		= rbyStyle ;
			register int					i ;

			/*	o^Ăꍇ̏B*/
			for (i = 0 ; i < MAX_SKKIME_COLORFACE ; i ++) {
				pColor->m_nTextColor		= skkimeConfig_isValidColor (*pByte)? *pByte : pDefault->m_nTextColor ;	pByte	++ ;
				pColor->m_nBackColor		= skkimeConfig_isValidColor (*pByte)? *pByte : pDefault->m_nBackColor ;	pByte	++ ;
				pColor->m_nUnderLineColor	= skkimeConfig_isValidColor (*pByte)? *pByte : pDefault->m_nUnderLineColor ;	pByte	++ ;
				pColor->m_nUnderLineType	= skkimeConfig_isValidLineType (*pByte)? *pByte : pDefault->m_nUnderLineType ;	pByte	++ ;
				pColor		++ ;
				pDefault	++ ;
			}
			fInitialized	= TRUE ;
		}
		MyRegCloseKey (hSubKey) ;
	}
	if (! fInitialized)
		memcpy (pColorFaceSet, pColorFaceSetDefault, sizeof (MYCOLORFACESET) * nColorFaceSetDefault) ;
	return	TRUE ;
}

BOOL
skkimeConfig_updateColorFace (
	register LPCMYSTR				strRegPath,
	register MYCOLORFACESET*		pColorFaceSet)
{
	HKEY						hSubKey ;
	BYTE						rbyStyle [MAX_SKKIME_COLORFACE * 4] ;
	register BOOL				fRetval	= TRUE ;
	register MYCOLORFACESET*	pColor ;
	register LPBYTE				pByte ;
	register int				i ;

	if (! skkimeConfig_CreateKey (strRegPath, FALSE, &hSubKey))
		return	FALSE ;

	pColor	= pColorFaceSet ;
	pByte	= rbyStyle ;
	for (i = 0 ; i < MAX_SKKIME_COLORFACE ; i ++) {
		*pByte	++	= (BYTE) pColor->m_nTextColor ;
		*pByte	++	= (BYTE) pColor->m_nBackColor ;
		*pByte	++	= (BYTE) pColor->m_nUnderLineColor ;
		*pByte	++	= (BYTE) pColor->m_nUnderLineType ;
		pColor	++ ;
	}
	if (MyRegSetValueEx (hSubKey, REGSUBKEY_COLORFACE, 0, REG_BINARY, rbyStyle, sizeof (rbyStyle)) != ERROR_SUCCESS) {
		fRetval	= FALSE ;
	}
	if (RegCloseKey (hSubKey) != ERROR_SUCCESS)
		fRetval	= FALSE ;
	return	fRetval ;
}

BOOL
skkimeConfig_isValidColor (
	register int		nColor)
{
	return	(0 <= nColor && nColor < MAX_MYCOLOR) ;
}

BOOL
skkimeConfig_isValidLineType (
	register int		nLineStyle)
{
	return	(0 <= nLineStyle && nLineStyle < MAX_MYLINE) ;
}






