/* # SKKIME (Simple Kana-Kanji Input Method Editor for Windows'98)
 * Config.c
 * This file is part of SKKIME.
 * Copyright (C) 1999
 * Takashi SAKAMOTO (tatari_sakamoto@nifty.ne.jp)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with skkinput; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "windows.h"
#include "tchar.h"
#include "immdev.h"
#include "commctrl.h"
#include "skki1_0.h"
#include <limits.h>
#include <stddef.h>
#include "conf\skimcnfp.h"
#include "resource.h"

#define	REGPATH_GENERIC	MYTEXT("Software\\TT\\Windows\\CurrentVersion\\SKKIME\\1.0\\Generic")
#define	REGSUBKEY_JHENKAN_SHOW_CANDIDATE_KEYS	MYTEXT("j-henkan-show-candidate-keys")
#define	REGSUBKEY_JINPUT_BY_CODE_OR_MENU_KEYS1	MYTEXT("j-input-by-code-or-menu-keys1")
#define	REGSUBKEY_JINPUT_BY_CODE_OR_MENU_KEYS2	MYTEXT("j-input-by-code-or-menu-keys2")

static	INT_PTR	skkimeConfig_dlgCandidateSelectKeyOnInitDialog (HWND, UINT, WPARAM, LPARAM) ;
static	INT_PTR	skkimeConfig_dlgCandidateSelectKeyOnNotify (HWND, UINT, WPARAM, LPARAM) ;
static	INT_PTR	skkimeConfig_dlgCandidateSelectKeyOnCommand (HWND, UINT, WPARAM, LPARAM) ;
static	BOOL	skkimeConfig_getCurrentCandidateSelectKeyFromDialog (HWND) ;

static	BYTE	srDefaultJHenkanShowCandidateKeys []	= {
	'a', 's', 'd', 'f', 'j', 'k', 'l'
} ;
static	BYTE	srDefaultJInputByCodeOrMenuKeys1 []		= {
	'a', 's', 'd', 'f', 'g', 'h', 'q', 'w', 'e', 'r', 't', 'y'
} ;
static	BYTE	srDefaultJInputByCodeOrMenuKeys2 []		= {
	'a', 's', 'd', 'f', 'g', 'h', 'j', 'k', 'l', 'q', 'w', 'e', 'r', 't', 'y', 'u'
} ;

BOOL
skkimeConfig_InitializeCandidateSelectKey (
	register TINYMEMMGR*		pMemMgr,
	register MYCANDSELKEY*		pArg)
{
	HKEY	hSubKey ;
	BOOL	fInitJHenkanShowCandidateKeys	= FALSE ;
	BOOL	fInitJInputByCodeOrMenuKeys1	= FALSE ;
	BOOL	fInitJInputByCodeOrMenuKeys2	= FALSE ;
	DWORD	dwType, cbData ;

	if (pMemMgr == NULL || pArg == NULL)
		return	FALSE ;

	if (MyRegOpenKeyEx (HKEY_CURRENT_USER, REGPATH_GENERIC, 0, KEY_READ, &hSubKey) == ERROR_SUCCESS) {
		if (MyRegQueryValueEx (hSubKey, REGSUBKEY_JHENKAN_SHOW_CANDIDATE_KEYS, NULL, &dwType, NULL, &cbData) == ERROR_SUCCESS &&
			dwType == REG_BINARY &&
			cbData == NUM_JHENKAN_SHOW_CANDIDATE_KEYS) {
			cbData	= NUM_JHENKAN_SHOW_CANDIDATE_KEYS ;
			(void) MyRegQueryValueEx (hSubKey, REGSUBKEY_JHENKAN_SHOW_CANDIDATE_KEYS, NULL, &dwType, pArg->_JHenkanShowCandidateKeys, &cbData) ;
			fInitJHenkanShowCandidateKeys	= TRUE ;
		}
		if (MyRegQueryValueEx (hSubKey, REGSUBKEY_JINPUT_BY_CODE_OR_MENU_KEYS1, NULL, &dwType, NULL, &cbData) == ERROR_SUCCESS &&
			dwType == REG_BINARY &&
			cbData == NUM_JINPUT_BY_CODE_OR_MENU_KEYS1) {
			cbData = NUM_JINPUT_BY_CODE_OR_MENU_KEYS1 ;
			(void) MyRegQueryValueEx (hSubKey, REGSUBKEY_JINPUT_BY_CODE_OR_MENU_KEYS1, NULL, &dwType, pArg->_JInputByCodeOrMenuKeys1, &cbData) ;
			fInitJInputByCodeOrMenuKeys1	= TRUE ;
		}
		if (MyRegQueryValueEx (hSubKey, REGSUBKEY_JINPUT_BY_CODE_OR_MENU_KEYS2, NULL, &dwType, NULL, &cbData) == ERROR_SUCCESS &&
			dwType == REG_BINARY &&
			cbData == NUM_JINPUT_BY_CODE_OR_MENU_KEYS2) {
			cbData = NUM_JINPUT_BY_CODE_OR_MENU_KEYS2 ;
			(void) MyRegQueryValueEx (hSubKey, REGSUBKEY_JINPUT_BY_CODE_OR_MENU_KEYS2, NULL, &dwType, pArg->_JInputByCodeOrMenuKeys2, &cbData) ;
			fInitJInputByCodeOrMenuKeys2	= TRUE ;
		}
		MyRegCloseKey (hSubKey) ;
	}
	if (! fInitJHenkanShowCandidateKeys)
		memcpy (pArg->_JHenkanShowCandidateKeys, srDefaultJHenkanShowCandidateKeys, NUM_JHENKAN_SHOW_CANDIDATE_KEYS) ;
	if (! fInitJInputByCodeOrMenuKeys1)
		memcpy (pArg->_JInputByCodeOrMenuKeys1, srDefaultJInputByCodeOrMenuKeys1, NUM_JINPUT_BY_CODE_OR_MENU_KEYS1) ;
	if (! fInitJInputByCodeOrMenuKeys2)
		memcpy (pArg->_JInputByCodeOrMenuKeys2, srDefaultJInputByCodeOrMenuKeys2, NUM_JINPUT_BY_CODE_OR_MENU_KEYS2) ;
	return	TRUE ;

	UNREFERENCED_PARAMETER (pMemMgr) ;
}

BOOL
skkimeConfig_UpdateCandidateSelectKey (
	register MYCANDSELKEY*	pArg)
{
	HKEY			hSubKey ;
	register BOOL	fRetval	= TRUE ;

	if (pArg == NULL)
		return	FALSE ;

	if (! skkimeConfig_CreateKey (REGPATH_GENERIC, FALSE, &hSubKey))
		return	FALSE ;

	if (MyRegSetValueEx (hSubKey, REGSUBKEY_JHENKAN_SHOW_CANDIDATE_KEYS, 0, REG_BINARY, (LPBYTE)pArg->_JHenkanShowCandidateKeys, NUM_JHENKAN_SHOW_CANDIDATE_KEYS) != ERROR_SUCCESS ||
		MyRegSetValueEx (hSubKey, REGSUBKEY_JINPUT_BY_CODE_OR_MENU_KEYS1, 0, REG_BINARY, (LPBYTE)pArg->_JInputByCodeOrMenuKeys1, NUM_JINPUT_BY_CODE_OR_MENU_KEYS1) != ERROR_SUCCESS ||
		MyRegSetValueEx (hSubKey, REGSUBKEY_JINPUT_BY_CODE_OR_MENU_KEYS2, 0, REG_BINARY, (LPBYTE)pArg->_JInputByCodeOrMenuKeys2, NUM_JINPUT_BY_CODE_OR_MENU_KEYS2) != ERROR_SUCCESS)
		fRetval	= FALSE ;
	
	if (RegCloseKey (hSubKey) != ERROR_SUCCESS)
		fRetval	= FALSE ;
	return	fRetval ;
}

INT_PTR	CALLBACK
skkimeConfig_dlgCandidateSelectKeyProc (
	register HWND		hDlg,
	register UINT		uMsg,
	register WPARAM		wParam,
	register LPARAM		lParam)
{
	switch (uMsg) {
	case	WM_INITDIALOG:
		return	skkimeConfig_dlgCandidateSelectKeyOnInitDialog (hDlg, uMsg, wParam, lParam) ;
	case	WM_NOTIFY:
		return	skkimeConfig_dlgCandidateSelectKeyOnNotify (hDlg, uMsg, wParam, lParam) ;
	case	WM_COMMAND:
		return	skkimeConfig_dlgCandidateSelectKeyOnCommand (hDlg, uMsg, wParam, lParam) ;
	default:
		break ;
	}
	return	FALSE ;
}

/*========================================================================*
 *	private functions
 */
INT_PTR
skkimeConfig_dlgCandidateSelectKeyOnInitDialog (
	register HWND		hDlg,
	register UINT		uMsg,
	register WPARAM		wParam,
	register LPARAM		lParam)
{
	register LPPROPSHEETPAGE	pPropSheet ;
	register SkkimeConfigArg*	pConfArg ;
	register MYCANDSELKEY*		pCandSel ;
	register HWND				hwndCB ;
	register LPBYTE				pBuffer ;
	register int				i, nRetval, nSel, nRes, ch ;
	MYCHAR						szBuffer [2] ;

	DEBUGPRINTF ((MYTEXT ("skkimeConfig_dlgCandidateSelectKeyOnInitDialog (%x)\n"), hDlg)) ;

	SetWindowLongPtr (hDlg, DWLP_USER, (LONG_PTR)lParam) ;
	pPropSheet	= (LPPROPSHEETPAGE) lParam ;
	pConfArg	= (SkkimeConfigArg *) pPropSheet->lParam ;
	pCandSel	= &pConfArg->m_CandSelKey ;

	/*	SĂ COMBOBOX Ƀhbv_EXg̗vfݒ肷B
	 */
	szBuffer [1]	= MYTEXT ('\0') ;
	for (nRes = IDC_COMBO_HCANDSEL1 ; nRes <= IDC_COMBO_JCODE1_12 ; nRes ++) {
		hwndCB	= GetDlgItem (hDlg, nRes) ;
		if (hwndCB == NULL)
			continue ;
		for (ch = 0 ; ch < (0x7F - 0x21) ; ch ++) {
			szBuffer [0]	= MYTEXT (' ') + 1 + ch ;
			MySendMessage (hwndCB, CB_INSERTSTRING, (WPARAM)ch, (LPARAM)szBuffer) ;
		}
	}

	/*	݂̐ݒ𔽉fB
	 */
	for (nRes = IDC_COMBO_HCANDSEL1 ; nRes <= IDC_COMBO_HCANDSEL7 ; nRes ++) {
		hwndCB	= GetDlgItem (hDlg, nRes) ;
		if (hwndCB == NULL)
			continue ;
		nSel	= (int) pCandSel->_JHenkanShowCandidateKeys [nRes - IDC_COMBO_HCANDSEL1] ;
		nSel	= (MYTEXT (' ') < nSel && nSel < 0x7F)? (nSel - 0x21) : -1 ;
		MySendMessage (hwndCB, CB_SETCURSEL, (WPARAM)nSel, (LPARAM) 0) ;
	}
	for (nRes = IDC_COMBO_JCODE1_1 ; nRes <= IDC_COMBO_JCODE1_12 ; nRes ++) {
		hwndCB	= GetDlgItem (hDlg, nRes) ;
		if (hwndCB == NULL)
			continue ;
		nSel	= (int) pCandSel->_JInputByCodeOrMenuKeys1 [nRes - IDC_COMBO_JCODE1_1] ;
		nSel	= (MYTEXT (' ') < nSel && nSel < 0x7F)? (nSel - 0x21) : -1 ;
		MySendMessage (hwndCB, CB_SETCURSEL, (WPARAM)nSel, (LPARAM) 0) ;
	}
	for (nRes = IDC_COMBO_JCODE2_1 ; nRes <= IDC_COMBO_JCODE2_16 ; nRes ++) {
		hwndCB	= GetDlgItem (hDlg, nRes) ;
		if (hwndCB == NULL)
			continue ;
		nSel	= (int) pCandSel->_JInputByCodeOrMenuKeys2 [nRes - IDC_COMBO_JCODE2_1] ;
		nSel	= (MYTEXT (' ') < nSel && nSel < 0x7F)? (nSel - 0x21) : -1 ;
		MySendMessage (hwndCB, CB_SETCURSEL, (WPARAM)nSel, (LPARAM) 0) ;
	}
	return	TRUE ;

	UNREFERENCED_PARAMETER (uMsg) ;
	UNREFERENCED_PARAMETER (wParam) ;
}

INT_PTR
skkimeConfig_dlgCandidateSelectKeyOnNotify (
	register HWND		hDlg,
	register UINT		uMsg,
	register WPARAM		wParam,
	register LPARAM		lParam)
{
	register LPPROPSHEETPAGE	pPropSheet ;
	register SkkimeConfigArg*	pConfArg ;
	register NMHDR	FAR*	pNMHDR	= (NMHDR *) lParam ;

	pPropSheet	= (LPPROPSHEETPAGE) GetWindowLongPtr (hDlg, DWLP_USER) ;
	pConfArg	= (SkkimeConfigArg *) pPropSheet->lParam ;

	switch (pNMHDR->code){
	case PSN_SETACTIVE:
		SetWindowLongPtr (hDlg, DWLP_MSGRESULT, 0) ;
		return	TRUE ;

	case PSN_KILLACTIVE:
		SetWindowLongPtr (hDlg, DWLP_MSGRESULT, FALSE) ;
		return	TRUE ;

	case PSN_APPLY:
	{
		if (skkimeConfig_getCurrentCandidateSelectKeyFromDialog (hDlg) &&
			skkimeConfig_UpdateCandidateSelectKey (&pConfArg->m_CandSelKey)) {
			SetWindowLongPtr (hDlg, DWLP_MSGRESULT, PSNRET_NOERROR) ;
		} else {
			SetWindowLongPtr (hDlg, DWLP_MSGRESULT, PSNRET_INVALID) ;
		}
		return	TRUE ;
	}
	case PSN_RESET:
	case PSN_HELP:
	default:
		return	FALSE ;
	}
	return	TRUE ;

	UNREFERENCED_PARAMETER (hDlg) ;
	UNREFERENCED_PARAMETER (uMsg) ;
	UNREFERENCED_PARAMETER (wParam) ;
}

INT_PTR
skkimeConfig_dlgCandidateSelectKeyOnCommand (
	register HWND		hDlg,
	register UINT		uMsg,
	register WPARAM		wParam,
	register LPARAM		lParam)
{
	register LPPROPSHEETPAGE	pPropSheet ;
	register SkkimeConfigArg*	pConfArg ;
	register WORD	wNotifyCode ;
	register WORD	wID ;
	register HWND	hwndCtl ;

	wNotifyCode	= HIWORD (wParam) ;
	wID			= LOWORD (wParam) ;
	hwndCtl		= (HWND) lParam ;

	pPropSheet	= (LPPROPSHEETPAGE) GetWindowLongPtr (hDlg, DWLP_USER) ;
	pConfArg	= (SkkimeConfigArg *) pPropSheet->lParam ;

	if (IDC_COMBO_HCANDSEL1 <= wID && wID <= IDC_COMBO_JCODE1_12) {
		if (wNotifyCode == CBN_SELCHANGE) 
			PropSheet_Changed (pConfArg->m_hPropSheetDlg, hDlg) ;
	}
	return	FALSE ;
	UNREFERENCED_PARAMETER (hDlg) ;
	UNREFERENCED_PARAMETER (uMsg) ;
	UNREFERENCED_PARAMETER (wParam) ;
	UNREFERENCED_PARAMETER (lParam) ;
}

/*	Dialog 猻݂ Generic Config ǂݎB
 *
 *	CheckBox ĂȂA̐l TRUE AȂ 
 *	FALSE AƂɂāB
 */
BOOL
skkimeConfig_getCurrentCandidateSelectKeyFromDialog (
	register HWND		hDlg)
{
	register LPPROPSHEETPAGE	pPropSheet ;
	register SkkimeConfigArg*	pConfArg ;
	register MYCANDSELKEY*		pCandSel ;
	register HWND				hwndCB ;
	register int				nRes, nSel ;

	pPropSheet	= (LPPROPSHEETPAGE) GetWindowLongPtr (hDlg, DWLP_USER) ;
	pConfArg	= (SkkimeConfigArg *) pPropSheet->lParam ;
	pCandSel	= &pConfArg->m_CandSelKey ;
	
	for (nRes = IDC_COMBO_HCANDSEL1 ; nRes <= IDC_COMBO_HCANDSEL7 ; nRes ++) {
		hwndCB	= GetDlgItem (hDlg, nRes) ;
		if (hwndCB == NULL)
			continue ;
		nSel	= MySendMessage (hwndCB, CB_GETCURSEL, (WPARAM)0, (LPARAM) 0) ;
		if (nSel != CB_ERR) {
			pCandSel->_JHenkanShowCandidateKeys [nRes - IDC_COMBO_HCANDSEL1]	= MYTEXT (' ') + 1 + nSel ;
		}
	}
	for (nRes = IDC_COMBO_JCODE1_1 ; nRes <= IDC_COMBO_JCODE1_12 ; nRes ++) {
		hwndCB	= GetDlgItem (hDlg, nRes) ;
		if (hwndCB == NULL)
			continue ;
		nSel	= MySendMessage (hwndCB, CB_GETCURSEL, (WPARAM)0, (LPARAM) 0) ;
		if (nSel != CB_ERR) {
			pCandSel->_JInputByCodeOrMenuKeys1 [nRes - IDC_COMBO_JCODE1_1]	= MYTEXT (' ') + 1 + nSel ;
		}
	}
	for (nRes = IDC_COMBO_JCODE2_1 ; nRes <= IDC_COMBO_JCODE2_16 ; nRes ++) {
		hwndCB	= GetDlgItem (hDlg, nRes) ;
		if (hwndCB == NULL)
			continue ;
		nSel	= MySendMessage (hwndCB, CB_GETCURSEL, (WPARAM)0, (LPARAM) 0) ;
		if (nSel != CB_ERR) {
			pCandSel->_JInputByCodeOrMenuKeys2 [nRes - IDC_COMBO_JCODE2_1]	= MYTEXT (' ') + 1 + nSel ;
		}
	}
	return	TRUE ;
}


