#include "windows.h"
#include "tchar.h"
#include "immdev.h"
#include "commctrl.h"
#include "commdlg.h"
#include "skki1_0.h"
#include <limits.h>
#include "conf\skimcnfp.h"
#include "TSearchSession.h"
#include "TJisyoUpdateSession.h"
#include "resource.h"

#define	REGPATH_DICTIONARY				MYTEXT("Software\\TT\\Windows\\CurrentVersion\\SKKIME\\1.0\\Dictionaries")
#define	REGKEY_USERJISYO				MYTEXT("UserJisyo")
#define	SKKIME_DEFAULT_JISYO_SETTING	MYTEXT("user=")
#if WINVER >= 0x0600
#define	SKKIME_DEFAULT_USERJISYOPATH	MYTEXT("%USERPROFILE%\\AppData\\Local\\TT\\skki1_0\\skki1_0u.dic")
#else
#define	SKKIME_DEFAULT_USERJISYOPATH	MYTEXT("%USERPROFILE%\\Application Data\\skk\\skki1_0\\skki1_0u.dic")
#endif

typedef struct {
	SkkimeConfigArg*	m_pConfArg ;
	LPPROPSHEETPAGE		m_pPropSheet ;
	HICON				m_hiconArrowUp ;
	HICON				m_hiconArrowDown ;
}	TDlgDictionaryParam ;

typedef struct {
	TINYMEMMGR*			m_pMemMgr ;
	HINSTANCE			m_hInst ;
	int					m_nType ;
	MYCHAR				m_strPath [MAX_PATH + 1] ;
	BOOL				m_fEnableAuxJisyo ;
	BOOL				m_fSortedAuxJisyo ;
	PORTANDHOSTNODE*	m_lstHost ;
}	TDlgEditDictArg ;

typedef struct {
	MYCHAR				m_strHost [MAX_PATH + 1] ;	// ?
	int					m_nPort ;
}	TDlgEditServerJisyoArg ;

static	INT_PTR	skkimeConfig_dlgDictionaryOnInitDialog	(HWND, UINT, WPARAM, LPARAM) ;
static	INT_PTR	skkimeConfig_dlgDictionaryOnNotify		(HWND, UINT, WPARAM, LPARAM) ;
static	INT_PTR	skkimeConfig_dlgDictionaryOnCommand		(HWND, UINT, WPARAM, LPARAM) ;
static	BOOL	skkimeConfig_editDictionary				(HWND, int, int) ;

static	INT_PTR	CALLBACK	skkimeConfig_dlgEditServerJisyo	(HWND, UINT, WPARAM, LPARAM) ;
static	INT_PTR	skkimeConfig_dlgEditServerJisyoOnInitDialog	(HWND, UINT, WPARAM, LPARAM) ;
static	INT_PTR	skkimeConfig_dlgEditServerJisyoOnCommand	(HWND, UINT, WPARAM, LPARAM) ;
static	INT_PTR	skkimeConfig_dlgEditServerJisyoOnNotify		(HWND, UINT, WPARAM, LPARAM) ;
static	BOOL	skkimeConfig_editServerJisyo				(HWND, int, int) ;
static	INT_PTR CALLBACK	skkimeConfig_dlgEditJisyoServer	(HWND, UINT, WPARAM, LPARAM) ;
static	INT_PTR	CALLBACK	skkimeConfig_dlgEditFileJisyo	(HWND, UINT, WPARAM, LPARAM) ;

static	BOOL				skkimeConfig_setDictionaryItemText	(HWND, int, DICTNODE*) ;
static	PORTANDHOSTNODE*	skkimeConfig_copyPortAndHostList	(TINYMEMMGR*, PORTANDHOSTNODE*) ;

static	BOOL	skkimeConfig_initSearchJisyoList (TINYMEMMGR*, LPCMYSTR, DICTNODE**, LPMYSTR) ;
static	BOOL	skkimeConfig_updateSearchJisyoList	(LPCMYSTR, DICTNODE*, LPMYSTR) ;
static	BOOL	skkimeConfig_parseDictNumber	(LPCMYSTR, int*) ;
static	DICTNODE*	skkimeConfig_parseDictionarySetting	(TINYMEMMGR*, LPCMYSTR, int) ;
static	PORTANDHOSTNODE*	skkimeConfig_parsePortAndHost	(TINYMEMMGR*, LPCMYSTR, int) ;
static	BOOL	skkimeConfig_insertDictionaryInSearchList	(DICTNODE**, DICTNODE*) ;
static	DICTNODE*	skkimeConfig_newDictionaryNode	(TINYMEMMGR*, int, LPCMYSTR) ;
static	BOOL	skkimeConfig_registerDictionaryNode	(DICTNODE**, DICTNODE*, DICTNODE*) ;
static	BOOL	skkimeConfig_unregisterDictionaryNode	(DICTNODE**, DICTNODE*) ;
static	void	skkimeConfig_deleteDictionaryNode	(TINYMEMMGR*, DICTNODE*) ;
static	void	skkimeConfig_deleteHostList	(TINYMEMMGR*, PORTANDHOSTNODE*) ;
static	PORTANDHOSTNODE*	skkimeConfig_newPortAndHostNode	(TINYMEMMGR*, LPCMYSTR, int, int) ;
static	BOOL	skkimeConfig_registerPortAndHostNode	(PORTANDHOSTNODE*, PORTANDHOSTNODE**, PORTANDHOSTNODE*) ;
static	BOOL	skkimeConfig_unregisterPortAndHostNode	(PORTANDHOSTNODE*, PORTANDHOSTNODE**) ;
static	BOOL	skkimeConfig_hostlist2string		(LPMYSTR, int, PORTANDHOSTNODE*) ;

static	BOOL	skkimeConfig_dlgDictionaryMoveSelectedNode	(HWND, SkkimeConfigArg*, BOOL) ;


/*========================================================================
 *	public functions
 */
BOOL
skkimeConfig_InitializeSearchJisyoList (
	register TINYMEMMGR*	pMemMgr,
	register DICTNODE**		plstJisyo,
	register LPMYSTR		strUserJisyoPath)
{
	if (pMemMgr == NULL || plstJisyo == NULL || strUserJisyoPath == NULL)
		return	FALSE ;
	return	skkimeConfig_initSearchJisyoList (pMemMgr, REGPATH_DICTIONARY, plstJisyo, strUserJisyoPath) ;
}

BOOL
skkimeConfig_UpdateSearchJisyoList (
	register DICTNODE*		lstJisyo,
	register LPMYSTR		strUserJisyoPath)
{
	if (lstJisyo == NULL || strUserJisyoPath == NULL)
		return	FALSE ;
	DEBUGPRINTF ((MYTEXT ("skkimeConfig_UpdateSearchJisyoList (%p, \"%s\")\n"),
				  lstJisyo, strUserJisyoPath)) ;
	return	skkimeConfig_updateSearchJisyoList (REGPATH_DICTIONARY, lstJisyo, strUserJisyoPath) ;
}

INT_PTR	CALLBACK
skkimeConfig_dlgDictionaryProc (
	register HWND		hDlg,
	register UINT		uMsg,
	register WPARAM		wParam,
	register LPARAM		lParam)
{
	switch (uMsg) {
	case	WM_INITDIALOG:
		return	skkimeConfig_dlgDictionaryOnInitDialog (hDlg, uMsg, wParam, lParam) ;
	case	WM_NOTIFY:
		return	skkimeConfig_dlgDictionaryOnNotify (hDlg, uMsg, wParam, lParam) ;
	case	WM_COMMAND:
		return	skkimeConfig_dlgDictionaryOnCommand (hDlg, uMsg, wParam, lParam) ;
	case	WM_DESTROY:
	{
		register TDlgDictionaryParam*	pParam ;
		register SkkimeConfigArg*		pConfArg ;
		register TINYMEMMGR*			pMemMgr ;

		pParam	= (TDlgDictionaryParam*) GetWindowLongPtr (hDlg, DWLP_USER) ;
		if (pParam == NULL)
			break ;

		pConfArg		= pParam->m_pConfArg ;
		pMemMgr			= &pConfArg->m_MemMgr ;
		if (pParam->m_hiconArrowUp != NULL) {
			DestroyIcon (pParam->m_hiconArrowUp) ;
			pParam->m_hiconArrowUp	= NULL ;
		}
		if (pParam->m_hiconArrowDown != NULL) {
			DestroyIcon (pParam->m_hiconArrowDown) ;
			pParam->m_hiconArrowDown	= NULL ;
		}
		TinyMalloc_Free (pMemMgr, pParam) ;
		SetWindowLongPtr (hDlg, DWLP_USER, 0) ;
		break ;
	}
	default:
		break ;
	}
	return	FALSE ;
}

/*========================================================================
 *	private functions
 */
INT_PTR
skkimeConfig_dlgDictionaryOnInitDialog (
	register HWND		hDlg,
	register UINT		uMsg,
	register WPARAM		wParam,
	register LPARAM		lParam)
{
	register LPPROPSHEETPAGE		pPropSheet ;
	register SkkimeConfigArg*		pConfArg ;
	register TINYMEMMGR*			pMemMgr ;
	register TDlgDictionaryParam*	pParam ;
	register HWND	hwndEditUsrDicPath ;
	register HWND	hwndListSearchDict ;
	register HWND	hwndButton ;

	pPropSheet		= (LPPROPSHEETPAGE) lParam ;
	pConfArg		= (SkkimeConfigArg *) pPropSheet->lParam ;
	pMemMgr			= &pConfArg->m_MemMgr ;

	pParam			= TinyMalloc_Malloc (pMemMgr, sizeof (TDlgDictionaryParam)) ;
	if (pParam == NULL)
		return	TRUE ;	/* Geeee... critical error... */

	pParam->m_pConfArg			= pConfArg ;
	pParam->m_pPropSheet		= pPropSheet ;
	pParam->m_hiconArrowUp		= NULL ;
	pParam->m_hiconArrowDown	= NULL ;
	SetWindowLongPtr (hDlg, DWLP_USER, (LONG_PTR) pParam) ;

	hwndEditUsrDicPath	= GetDlgItem (hDlg, IDC_EDIT_USRDICPATH) ;
	if (hwndEditUsrDicPath != NULL) {
		MYCHAR	szPath [MAX_PATH + 1] ;
		if (! MyExpandEnvironmentStrings (pConfArg->m_rszUserJisyoPath, szPath, MAX_PATH)) {
			MySendMessage (hwndEditUsrDicPath, WM_SETTEXT, 0, (LPARAM) pConfArg->m_rszUserJisyoPath) ;
		} else {
			MySendMessage (hwndEditUsrDicPath, WM_SETTEXT, 0, (LPARAM) szPath) ;
		}
		MySendMessage (hwndEditUsrDicPath, EM_SETLIMITTEXT, (WPARAM) MAX_PATH, 0) ;
	}

	hwndListSearchDict	= GetDlgItem (hDlg, IDC_LIST_SEARCHDICT) ;
	if (hwndListSearchDict != NULL) {
		register DICTNODE*	pNode ;
		MYLV_COLUMN			lvColumn;
		MYLVITEM			lvI ;
		register int		n ;
		static LPMYSTR	rpszDicType []	= {
			MYTEXT ("[U"),		MYTEXT ("T[o"),
			MYTEXT ("t@C"),	MYTEXT ("\[gσt@C"),
		} ;
		
		MyListView_DeleteAllItems (hwndListSearchDict) ;
		MyListView_SetExtendedListViewStyle (hwndListSearchDict, LVS_EX_FULLROWSELECT) ;
		lvColumn.mask		= LVCF_FMT | LVCF_WIDTH | LVCF_TEXT | LVCF_SUBITEM ;
		lvColumn.fmt		= LVCFMT_LEFT ;
		lvColumn.cx			= 120 ;
		lvColumn.pszText	= MYTEXT ("") ;
		MyListView_InsertColumn (hwndListSearchDict, 0, &lvColumn) ;
		lvColumn.cx			= 240 ;
		lvColumn.pszText	= MYTEXT ("ݒ") ;
		MyListView_InsertColumn (hwndListSearchDict, 1, &lvColumn) ;
		
		lvI.mask		= LVIF_PARAM | LVIF_STATE ;
		lvI.state		= 0 ; 
		lvI.stateMask	= 0 ;
		lvI.iSubItem	= 0 ;
		lvI.pszText		= 0 ;
		pNode	= pConfArg->m_lstSearchJisyo ;
		n		= 0 ;
		while (pNode != NULL) {
			register int	nItem ;
			lvI.iItem		= n ;
			lvI.iImage		= n ;
			lvI.lParam		= (LPARAM) pNode ;
			nItem			= MyListView_InsertItem (hwndListSearchDict, &lvI) ;
			if (nItem != -1) 
				skkimeConfig_setDictionaryItemText (hwndListSearchDict, nItem, pNode) ;
			pNode	= pNode->m_pNext ;
			n		++ ;
		}
	}
	hwndButton	= GetDlgItem (hDlg, IDC_BUTTON_ARROWUP) ;
	if (hwndButton != NULL) {
		register HICON	hIcon ;
		hIcon	= LoadImage (g_hInst, TEXT ("INDICICONL"), IMAGE_ICON, 16, 16, LR_SHARED) ;
		if (hIcon != NULL) {
			SendMessage (hwndButton, BM_SETIMAGE, (WPARAM) IMAGE_ICON, (LPARAM) hIcon) ;
			pParam->m_hiconArrowUp		= hIcon ;
		}
	}
	hwndButton	= GetDlgItem (hDlg, IDC_BUTTON_ARROWDOWN) ;
	if (hwndButton != NULL) {
		register HICON	hIcon ;
		hIcon	= LoadImage (g_hInst, TEXT ("INDICICONK"), IMAGE_ICON, 16, 16, LR_SHARED) ;
		if (hIcon != NULL) {
			SendMessage (hwndButton, BM_SETIMAGE, (WPARAM) IMAGE_ICON, (LPARAM) hIcon) ;
			pParam->m_hiconArrowDown	= hIcon ;
		}
	}
	return	TRUE ;
	UNREFERENCED_PARAMETER (uMsg) ;
	UNREFERENCED_PARAMETER (wParam) ;
}

INT_PTR
skkimeConfig_dlgDictionaryOnNotify (
	register HWND		hDlg,
	register UINT		uMsg,
	register WPARAM		wParam,
	register LPARAM		lParam)
{
	register SkkimeConfigArg*		pConfArg ;
	register TDlgDictionaryParam*	pParam ;
	register NMHDR	FAR*	pNMHDR	= (NMHDR *) lParam ;
	static TMYMENUITEM		rmi []	= {
		{ MIIM_ID | MIIM_STRING, 0,  TEXT("ǉ"),		IDC_BUTTON_ADD_FILE_JISYO, },
		{ MIIM_ID | MIIM_STRING, 0,  TEXT("T[oǉ"),IDC_BUTTON_ADD_SERVER_JISYO, },
		{ MIIM_ID | MIIM_STRING, 0,  TEXT("ݒύX"),	IDC_BUTTON_EDIT_DICT_IN_SEARCHLIST, },
		{ MIIM_ID | MIIM_STRING, 0,  TEXT("폜"),		IDC_BUTTON_REMOVE_DICT_FROM_SEARCHLIST, },
	} ;

	pParam	= (TDlgDictionaryParam*) GetWindowLongPtr (hDlg, DWLP_USER) ;
	if (pParam == NULL)
		return	1 ;

	pConfArg	= pParam->m_pConfArg ;

	switch (pNMHDR->idFrom) {
	case	IDC_LIST_SEARCHDICT:
	{
		register HWND			hwndLV	= GetDlgItem (hDlg, IDC_LIST_SEARCHDICT) ;
		register NMLISTVIEW*	pNM		= (NMLISTVIEW*) lParam ;
		register int			n ;

		switch (pNMHDR->code) {
		case	NM_RCLICK:
			n		= skkimeConfig_PopupMenu (hDlg, rmi, ARRAYSIZE (rmi)) ;
			if (n > 0) 
				skkimeConfig_editDictionary (hDlg, pNM->iItem, n) ;
			break ;
		default:
			break ;
		}
		break ;
	}
	default:
		switch (pNMHDR->code){
		case PSN_SETACTIVE:
		case PSN_KILLACTIVE:
			SetWindowLongPtr (hDlg, DWLP_MSGRESULT, 0) ;
			return	TRUE ;

		case PSN_APPLY:
		{
			if (skkimeConfig_UpdateSearchJisyoList (pConfArg->m_lstSearchJisyo, pConfArg->m_rszUserJisyoPath)) {
				SetWindowLongPtr (hDlg, DWLP_MSGRESULT, PSNRET_NOERROR) ;
			} else {
				SetWindowLongPtr (hDlg, DWLP_MSGRESULT, PSNRET_INVALID) ;
			}
			return	TRUE ;
		}
		case PSN_RESET:
		case PSN_HELP:
			break ;
		default:
			return FALSE ;
		}
	}
	return	TRUE ;
	
	UNREFERENCED_PARAMETER (uMsg) ;
	UNREFERENCED_PARAMETER (wParam) ;
}

INT_PTR
skkimeConfig_dlgDictionaryOnCommand (
	register HWND		hDlg,
	register UINT		uMsg,
	register WPARAM		wParam,
	register LPARAM		lParam)
{
	register TDlgDictionaryParam*	pParam ;
	register SkkimeConfigArg*		pConfArg ;
	register TINYMEMMGR*			pMemMgr ;
	register WORD				wNotifyCode ;
	register WORD				wID ;
	register HWND				hwndCtl ;

	pParam		= (TDlgDictionaryParam*) GetWindowLongPtr (hDlg, DWLP_USER) ;
	wNotifyCode	= HIWORD (wParam) ;
	wID			= LOWORD (wParam) ;
	hwndCtl		= (HWND) lParam ;
	pConfArg	= pParam->m_pConfArg ;
	pMemMgr		= &pConfArg->m_MemMgr ;

	switch (wID) {
	case	IDC_EDIT_USRDICPATH:
		/*	GfBbg{bNXҏWƁAAPPLY LɂȂȂ΁B*/
		if (wNotifyCode == EN_CHANGE) 
			PropSheet_Changed (pConfArg->m_hPropSheetDlg, hDlg) ;
		break ;
	case	IDC_BUTTON_BROWSE_USRDICTPATH:
	{
		register SkkimeConfigArg*	pConfArg	= pParam->m_pConfArg ;
		MYOPENFILENAME	ofn ;
		MYCHAR			szPath [MAX_PATH + 1] ;
		
		if (wNotifyCode == BN_CLICKED) {
			if (! MyExpandEnvironmentStrings (pConfArg->m_rszUserJisyoPath, szPath, MAX_PATH)) 
				Mylstrcpyn (szPath, pConfArg->m_rszUserJisyoPath, MAX_PATH) ;
			szPath [MAX_PATH]	= MYTEXT ('\0') ;
			memset (&ofn, 0, sizeof (ofn)) ;
			ofn.lStructSize			= sizeof (MYOPENFILENAME) ;
			ofn.hwndOwner			= hDlg ;
			ofn.hInstance			= pConfArg->m_hInst ;
			ofn.lpstrFilter			= MYTEXT ("All\0*.*\0") ;
			ofn.lpstrCustomFilter	= NULL ;
			ofn.nFilterIndex		= 1 ;
			ofn.lpstrFile			= szPath ;
			ofn.nMaxFile			= MAX_PATH ;
			ofn.lpstrFileTitle		= NULL ;
			ofn.lpstrInitialDir		= MYTEXT ("") ;
			ofn.lpstrTitle			= MYTEXT ("[ȖI") ;
			ofn.Flags				= OFN_CREATEPROMPT | OFN_HIDEREADONLY | OFN_LONGNAMES | OFN_NONETWORKBUTTON ;
			ofn.lpstrDefExt			= NULL ;
			ofn.lpfnHook			= NULL ;
			if (MyGetOpenFileName (&ofn) != 0) {
				Mylstrcpyn (pConfArg->m_rszUserJisyoPath, szPath, MAX_PATH) ;
				pConfArg->m_rszUserJisyoPath [MAX_PATH]	= MYTEXT ('\0') ;
				MySetDlgItemText (hDlg, IDC_EDIT_USRDICPATH, szPath) ;
				PropSheet_Changed (pConfArg->m_hPropSheetDlg, hDlg) ;
			}
			return	0 ;
		}
		return	1 ;
	}
	case	IDC_BUTTON_SAVEUSERJISYO:
	{
		/*	skkiserv Ɏ̕ۑvB*/
		TJisyoSaveSession_bSynchronize () ;
		return	1 ;
	}
	case	IDC_BUTTON_REPAIR_USRDICT:
		return	1 ;
	case	IDC_BUTTON_EDIT_DICT_IN_SEARCHLIST:
	case	IDC_BUTTON_ADD_FILE_JISYO:
	case	IDC_BUTTON_ADD_SERVER_JISYO:
	{
		if (wNotifyCode == BN_CLICKED) {
			register HWND	hwndLV	= GetDlgItem (hDlg, IDC_LIST_SEARCHDICT) ;
			register int	nItem ;

			if (hwndLV == NULL)
				return	1 ;
			nItem	= MyListView_GetSelectionMark (hwndLV) ;
			skkimeConfig_editDictionary (hDlg, nItem, wID) ;
			return	0 ;
		}
		return	1 ;
	}
	case	IDC_BUTTON_REMOVE_DICT_FROM_SEARCHLIST:
	{
		if (wNotifyCode == BN_CLICKED) {
			register HWND		hwndLV		= GetDlgItem (hDlg, IDC_LIST_SEARCHDICT) ;
			register DICTNODE**	plstJisyo	= &pConfArg->m_lstSearchJisyo ;
			register int		nItem ;
			register DICTNODE*	pNode ;
			MYLVITEM			lvI ;

			if (hwndLV == NULL)
				return	1 ;

			nItem			= MyListView_GetSelectionMark (hwndLV) ;
			if (nItem < 0)
				return	1 ;
			lvI.mask		= LVIF_PARAM ;
			lvI.iItem		= nItem ;
			if (! MyListView_GetItem (hwndLV, &lvI)) {
				pNode		= NULL ;
			} else {
				pNode		= (DICTNODE *)lvI.lParam ;
			}
			if (pNode == NULL)
				return	1 ;

			MyListView_DeleteItem (hwndLV, nItem) ;
			skkimeConfig_unregisterDictionaryNode (plstJisyo, pNode) ;
			skkimeConfig_deleteDictionaryNode (pMemMgr, pNode) ;
			PropSheet_Changed (pConfArg->m_hPropSheetDlg, hDlg) ;
			return	0 ;
		}
	}
	case	IDC_BUTTON_ARROWUP:
		if (wNotifyCode == BN_CLICKED) {
			if (skkimeConfig_dlgDictionaryMoveSelectedNode (hDlg, pConfArg, TRUE))
				PropSheet_Changed (pConfArg->m_hPropSheetDlg, hDlg) ;
			return	0 ;
		}
		break ;
	case	IDC_BUTTON_ARROWDOWN:
		if (wNotifyCode == BN_CLICKED) {
			if (skkimeConfig_dlgDictionaryMoveSelectedNode (hDlg, pConfArg, FALSE))
				PropSheet_Changed (pConfArg->m_hPropSheetDlg, hDlg) ;
			return	0 ;
		}
		break ;
	default:
		break ;
	}
	return	1 ;
	UNREFERENCED_PARAMETER (uMsg) ;
}

BOOL
skkimeConfig_editDictionary (
	register HWND		hDlg,
	register int		nItem,
	register int		nCommand)
{
	register TDlgDictionaryParam*	pParam ;
	register SkkimeConfigArg*	pConfArg ;
	register HWND				hwndLV ;
	register TINYMEMMGR*		pMemMgr ;
	register HINSTANCE			hInst ;
	register DICTNODE**			plstJisyo ;
	register DICTNODE*			pNode ;
	register DICTNODE*			pNewNode ;
	TDlgEditDictArg				arg ;
	MYLVITEM					lvI ;
	register int				nRetval, nPath ;

	pParam		= (TDlgDictionaryParam *) GetWindowLongPtr (hDlg, DWLP_USER) ;
	hwndLV		= GetDlgItem (hDlg, IDC_LIST_SEARCHDICT) ;
	if (hwndLV == NULL || pParam == NULL || pParam->m_pConfArg == NULL)
		return	FALSE ;
	pConfArg	= pParam->m_pConfArg ;
	pMemMgr		= &pConfArg->m_MemMgr ;
	plstJisyo	= &pConfArg->m_lstSearchJisyo ;
	hInst		= pConfArg->m_hInst ;

	memset (&lvI, 0, sizeof (lvI)) ;
	lvI.mask		= LVIF_PARAM ;
	lvI.iItem		= nItem ;
	if (! MyListView_GetItem (hwndLV, &lvI)) {
		pNode		= NULL ;
	} else {
		pNode		= (DICTNODE *)lvI.lParam ;
	}

	arg.m_hInst			= hInst ;
	arg.m_pMemMgr		= pMemMgr ;
	arg.m_strPath [0]	= TEXT ('\0') ;
	arg.m_lstHost		= NULL ;

	switch (nCommand) {
	case	IDC_BUTTON_EDIT_DICT_IN_SEARCHLIST:
	{
		if (pNode == NULL)
			return	FALSE ;
		if (pNode->m_nType == DICTYPE_USER)
			return	TRUE ;

		arg.m_nType		= pNode->m_nType ;
		arg.m_lstHost	= pNode->m_lstHost ;
		if (pNode->m_strPath != NULL) {
			Mylstrcpy (arg.m_strPath, pNode->m_strPath) ;
		} else {
			arg.m_strPath [0]	= MYTEXT ('\0') ;
		}
		Mylstrcpy (arg.m_strPath, pNode->m_strPath) ;
		arg.m_fEnableAuxJisyo	= pNode->m_fEnableAuxJisyo ;
		arg.m_fSortedAuxJisyo	= pNode->m_fSortedAuxJisyo ;

		if (pNode->m_nType == DICTYPE_SERVER) {
			nRetval	= DialogBoxParam (hInst, MAKEINTRESOURCE (IDD_DIALOG_EDIT_SERVER_JISYO), hDlg, skkimeConfig_dlgEditServerJisyo, (LPARAM) &arg) ;
		} else if (pNode->m_nType == DICTYPE_FILE || pNode->m_nType == DICTYPE_SORT) {
			nRetval	= DialogBoxParam (hInst, MAKEINTRESOURCE (IDD_DIALOG_EDIT_FILE_JISYO), hDlg, skkimeConfig_dlgEditFileJisyo, (LPARAM) &arg) ;
		} else {
			return	FALSE ;	/* unknown jisyo type */
		}
		break ;
	}
	case	IDC_BUTTON_ADD_FILE_JISYO:
	{
		arg.m_nType		= DICTYPE_FILE ;
		nRetval	= DialogBoxParam (hInst, MAKEINTRESOURCE (IDD_DIALOG_EDIT_FILE_JISYO), hDlg, skkimeConfig_dlgEditFileJisyo, (LPARAM) &arg) ;
		break ;
	}
	case	IDC_BUTTON_ADD_SERVER_JISYO:
	{
		arg.m_nType				= DICTYPE_SERVER ;
		arg.m_fEnableAuxJisyo	= FALSE ;
		arg.m_fSortedAuxJisyo	= FALSE ;
		nRetval	= DialogBoxParam (hInst, MAKEINTRESOURCE (IDD_DIALOG_EDIT_SERVER_JISYO), hDlg, skkimeConfig_dlgEditServerJisyo, (LPARAM) &arg) ;
		break ;
	}
	case	IDC_BUTTON_REMOVE_DICT_FROM_SEARCHLIST:
	{
		if (pNode == NULL || pNode->m_nType == DICTYPE_USER)
			return	FALSE ;
		MyListView_DeleteItem (hwndLV, nItem) ;
		skkimeConfig_unregisterDictionaryNode (plstJisyo, pNode) ;
		skkimeConfig_deleteDictionaryNode (pMemMgr, pNode) ;
		return	TRUE ;
	}
	default:
		return	FALSE ;
	}

	/*	nRetval ͂ēB
	 */
	if (nRetval != IDOK) 
		return	TRUE ;
	switch (arg.m_nType) {
	case	DICTYPE_SERVER:
		if (arg.m_lstHost == NULL) {
			DEBUGPRINTF ((TEXT ("server-jisyo: no host/port\n"))) ;
			return	FALSE ;
		}
		break ;
	case	DICTYPE_SORT:
	case	DICTYPE_FILE:
		if (arg.m_strPath [0] == TEXT ('\0')) {
			DEBUGPRINTF ((TEXT ("file- or sort-jisyo: no path\n"))) ;
			return	FALSE ;
		}
		break ;
	default:
		return	FALSE ;
	}

	nPath		= Mylstrlen (arg.m_strPath) ;
	pNewNode	= skkimeConfig_newDictionaryNode (pMemMgr, arg.m_nType, arg.m_strPath) ;
	if (pNewNode == NULL) {
		if (pNode->m_lstHost != arg.m_lstHost && arg.m_lstHost != NULL) 
			skkimeConfig_deleteHostList (pMemMgr, arg.m_lstHost) ;
		return	FALSE ;	// out of memory 
	}
	if (arg.m_nType == DICTYPE_SERVER) {
		pNewNode->m_fSortedAuxJisyo	= arg.m_fSortedAuxJisyo ;
		pNewNode->m_fEnableAuxJisyo	= arg.m_fEnableAuxJisyo ;

		pNewNode->m_lstHost			= arg.m_lstHost ;
		if (pNode != NULL && pNode->m_lstHost == arg.m_lstHost) 
			pNode->m_lstHost	= NULL ;
		arg.m_lstHost		= NULL ;
	} else if (arg.m_nType == DICTYPE_FILE || arg.m_nType == DICTYPE_SORT) {
		pNewNode->m_fSortedAuxJisyo	= FALSE ;
		pNewNode->m_fEnableAuxJisyo	= FALSE ;

		pNewNode->m_lstHost	= NULL ;
	}

	memset (&lvI, 0, sizeof (lvI)) ;
	lvI.mask		= LVIF_PARAM | LVIF_STATE ;
	lvI.lParam		= (LPARAM) pNewNode ;
	if (nCommand == IDC_BUTTON_EDIT_DICT_IN_SEARCHLIST) {
		lvI.iItem	= nItem ;
		MyListView_SetItem (hwndLV, &lvI) ;
		skkimeConfig_registerDictionaryNode (plstJisyo, pNewNode, pNode) ;
		skkimeConfig_unregisterDictionaryNode (plstJisyo, pNode) ;
		skkimeConfig_deleteDictionaryNode (pMemMgr, pNode) ;
		skkimeConfig_setDictionaryItemText (hwndLV, nItem, pNewNode) ;
	} else {
		skkimeConfig_registerDictionaryNode (plstJisyo, pNewNode, pNode) ;
		lvI.iItem	= (pNode == NULL)? 0 : (nItem + 1) ;
		nItem		= MyListView_InsertItem (hwndLV, &lvI) ;
		if (nItem != -1) 
			skkimeConfig_setDictionaryItemText (hwndLV, nItem, pNewNode) ;
	}
	PropSheet_Changed (pConfArg->m_hPropSheetDlg, hDlg) ;
	return	TRUE ;
}

INT_PTR	CALLBACK
skkimeConfig_dlgEditServerJisyo (
	register HWND		hDlg,
	register UINT		uMsg,
	register WPARAM		wParam,
	register LPARAM		lParam)
{
	switch (uMsg) {
	case	WM_INITDIALOG:
		return	skkimeConfig_dlgEditServerJisyoOnInitDialog (hDlg, uMsg, wParam, lParam) ;
	case	WM_COMMAND:
		return	skkimeConfig_dlgEditServerJisyoOnCommand (hDlg, uMsg, wParam, lParam) ;
	case	WM_NOTIFY:
		return	skkimeConfig_dlgEditServerJisyoOnNotify (hDlg, uMsg, wParam, lParam) ;
	default:
		break ;
	}
	return	FALSE ;
}

INT_PTR
skkimeConfig_dlgEditServerJisyoOnInitDialog	(
	register HWND		hDlg,
	register UINT		uMsg,
	register WPARAM		wParam,
	register LPARAM		lParam)
{
	register PORTANDHOSTNODE*	pNode ;
	register TDlgEditDictArg*	pArg ;
	register TINYMEMMGR*		pMemMgr ;
	register int				nItem, n ;
	MYLV_COLUMN					lvColumn;
	MYLVITEM					lvI ;
	register HWND				hwndAuxPath, hwndCheck, hwndLV ;
	MYCHAR						szBuffer [32] ;

	SetWindowLongPtr (hDlg, DWLP_USER, (LONG_PTR) lParam) ;
	pArg			= (TDlgEditDictArg*) lParam ;
	if (pArg == NULL)
		return	TRUE ;
	pMemMgr			= pArg->m_pMemMgr ;

	hwndAuxPath		= GetDlgItem (hDlg, IDC_EDIT_FILE_DICT_PATH) ;
	if (hwndAuxPath != NULL) {
		MySendMessage   (hwndAuxPath, EM_LIMITTEXT, (WPARAM) MAX_PATH, 0) ;
		MySetWindowText (hwndAuxPath, pArg->m_strPath) ;
	}

	/*	XgRs[BIW𑀍삷킯ɂ͂Ȃ̂ŁB*/
	pArg->m_lstHost	= skkimeConfig_copyPortAndHostList (pMemMgr, pArg->m_lstHost) ;
	hwndLV			= GetDlgItem (hDlg, IDC_LIST_SERVER_DICT) ;

	if (hwndLV != NULL) {
		MyListView_DeleteAllItems (hwndLV) ;
		MyListView_SetExtendedListViewStyle (hwndLV, LVS_EX_FULLROWSELECT) ;
	
		lvColumn.mask		= LVCF_FMT | LVCF_WIDTH | LVCF_TEXT | LVCF_SUBITEM ;
		lvColumn.fmt		= LVCFMT_LEFT ;
		lvColumn.pszText	= MYTEXT ("zXg") ;
		lvColumn.cx			= 120 ;
		MyListView_InsertColumn (hwndLV, 0, &lvColumn) ;
		lvColumn.pszText	= MYTEXT ("|[gԍ") ;
		MyListView_InsertColumn (hwndLV, 1, &lvColumn) ;
		
		pNode			= pArg->m_lstHost ;
		memset (&lvI, 0, sizeof (lvI)) ;
		lvI.mask		= LVIF_TEXT | LVIF_PARAM | LVIF_STATE ;
		n				= 0 ;
		while (pNode != NULL) {
			lvI.iItem	= n ;
			lvI.pszText	= pNode->m_strHost ;
			lvI.lParam	= (LPARAM) pNode ;
			nItem		= MyListView_InsertItem (hwndLV, &lvI) ;
			if (nItem != -1) {
				Mywsprintf (szBuffer, MYTEXT ("%d"), pNode->m_nPort) ;
				MyListView_SetItemText (hwndLV, nItem, 1, szBuffer) ;
			}
			pNode		= pNode->m_pNext ;
			n			++ ;
		}
	}

	CheckDlgButton (hDlg, IDC_CHECK_FILE_DICT_SORTEDP, (pArg->m_fSortedAuxJisyo)? BST_CHECKED : BST_UNCHECKED) ;
	CheckDlgButton (hDlg, IDC_CHECK_ENABLE_AUX_JISYO,  (pArg->m_fEnableAuxJisyo)? BST_CHECKED : BST_UNCHECKED) ;
	if (hwndAuxPath != NULL)
		EnableWindow (hwndAuxPath, pArg->m_fEnableAuxJisyo) ;
	hwndCheck	= GetDlgItem (hDlg, IDC_CHECK_FILE_DICT_SORTEDP) ;
	if (hwndCheck != NULL)
		EnableWindow (hwndCheck, pArg->m_fEnableAuxJisyo) ;
	
	return	TRUE ;
	UNREFERENCED_PARAMETER (uMsg) ;
	UNREFERENCED_PARAMETER (wParam) ;
}

INT_PTR
skkimeConfig_dlgEditServerJisyoOnCommand (
	register HWND		hDlg,
	register UINT		uMsg,
	register WPARAM		wParam,
	register LPARAM		lParam)
{
	register WORD				wNotifyCode ;
	register WORD				wID ;
	register HWND				hwndCtl ;
	register TDlgEditDictArg*	pArg ;
	register TINYMEMMGR*		pMemMgr ;

	pArg		= (TDlgEditDictArg*) GetWindowLongPtr (hDlg, DWLP_USER) ;
	pMemMgr		= pArg->m_pMemMgr ;
	wNotifyCode	= HIWORD (wParam) ;
	wID			= LOWORD (wParam) ;
	hwndCtl		= (HWND) lParam ;

	switch (wID) {
	case	IDC_BUTTON_ADD_NEW_JISYO_SERVER:
	case	IDC_BUTTON_MODIFY_JISYO_SERVER:
	case	IDC_BUTTON_REMOVE_JISYO_SERVER:
	{
		if (wNotifyCode == BN_CLICKED) {
			register HWND	hwndLV	= GetDlgItem (hDlg, IDC_LIST_SERVER_DICT) ;
			register int	nItem ;

			if (hwndLV == NULL)
				break ;
			nItem	= MyListView_GetSelectionMark (hwndLV) ;
			skkimeConfig_editServerJisyo (hDlg, nItem, wID) ;
			return	FALSE ;
		}
		break ;
	}
	case	IDC_CHECK_FILE_DICT_SORTEDP:
	{
		if (wNotifyCode == BN_CLICKED) {
			register UINT	uCheck ;
			register BOOL	fEnable ;

			uCheck		= IsDlgButtonChecked (hDlg, IDC_CHECK_FILE_DICT_SORTEDP) ;
			pArg->m_fSortedAuxJisyo	= (uCheck == BST_CHECKED)? TRUE : FALSE ;
			return	FALSE ;
		}
	}
	case	IDC_CHECK_ENABLE_AUX_JISYO:
	{
		if (wNotifyCode == BN_CLICKED) {
			register UINT	uCheck ;
			register BOOL	fEnable ;

			uCheck		= IsDlgButtonChecked (hDlg, IDC_CHECK_ENABLE_AUX_JISYO) ;
			fEnable		= (uCheck == BST_CHECKED)? TRUE : FALSE ;
			if (fEnable != pArg->m_fEnableAuxJisyo) {
				register HWND	hwndCheck, hwndEdit ;

				hwndEdit	= GetDlgItem (hDlg, IDC_EDIT_FILE_DICT_PATH) ;
				hwndCheck	= GetDlgItem (hDlg, IDC_CHECK_FILE_DICT_SORTEDP) ;
				if (hwndEdit != NULL)
					EnableWindow (hwndEdit,  fEnable) ;
				if (hwndCheck != NULL)
					EnableWindow (hwndCheck, fEnable) ;
				pArg->m_fEnableAuxJisyo	= fEnable ;
			}
			return	FALSE ;
		}
		break ;
	}
	case	IDC_BUTTON_BROWSE_FILE_DICT_PATH:
	{
		MYOPENFILENAME	ofn ;
		MYCHAR			szPath [MAX_PATH + 1] ;

		if (wNotifyCode == BN_CLICKED) {
			Mylstrcpyn (szPath, pArg->m_strPath, MAX_PATH) ;
			szPath [MAX_PATH]	= MYTEXT ('\0') ;
			memset (&ofn, 0, sizeof (ofn)) ;
			ofn.lStructSize			= sizeof (MYOPENFILENAME) ;
			ofn.hwndOwner			= hDlg ;
			ofn.hInstance			= pArg->m_hInst ;
			ofn.lpstrFilter			= MYTEXT ("All\0*.*\0") ;
			ofn.lpstrCustomFilter	= NULL ;
			ofn.nFilterIndex		= 1 ;
			ofn.lpstrFile			= szPath ;
			ofn.nMaxFile			= MAX_PATH ;
			ofn.lpstrFileTitle		= NULL ;
			ofn.lpstrInitialDir		= MYTEXT ("") ;
			ofn.lpstrTitle			= MYTEXT ("⏕̑I") ;
			ofn.Flags				= OFN_HIDEREADONLY | OFN_LONGNAMES | OFN_NONETWORKBUTTON | OFN_PATHMUSTEXIST ;
			ofn.lpstrDefExt			= NULL ;
			ofn.lpfnHook			= NULL ;
			if (MyGetOpenFileName (&ofn) != 0) {
				Mylstrcpyn (pArg->m_strPath, szPath, MAX_PATH) ;
				pArg->m_strPath [MAX_PATH]	= MYTEXT ('\0') ;
				MySetDlgItemText (hDlg, IDC_EDIT_FILE_DICT_PATH, szPath) ;
			}
			return	FALSE ;
		}
		break ;
	}
	case	IDOK:
	case	IDCANCEL:
	{
		if (wID == IDOK) {
			if (pArg != NULL) {
				register HWND	hwndAuxPath ;

				hwndAuxPath		= GetDlgItem (hDlg, IDC_EDIT_FILE_DICT_PATH) ;
				pArg->m_strPath [0]	= TEXT ('\0') ;
				if (hwndAuxPath != NULL) 
					MyGetWindowText (hwndAuxPath, pArg->m_strPath, MAX_PATH) ;

				pArg->m_fSortedAuxJisyo	= (IsDlgButtonChecked (hDlg, IDC_CHECK_FILE_DICT_SORTEDP) == BST_CHECKED)? TRUE : FALSE ;
				
				/*	host-port ̗́ceiXĂ̂
				 *	ŔoKv͂ȂB*/
			} else {
				wID	= IDCANCEL ;
			}
		}
		if (wID == IDCANCEL) {
			skkimeConfig_deleteHostList (pMemMgr, pArg->m_lstHost) ;
			pArg->m_lstHost	= NULL ;
		}
		EndDialog (hDlg, wID) ;
		return	FALSE ;
	}
	default:
		break ;
	}
	return	TRUE ;
	UNREFERENCED_PARAMETER (uMsg) ;
}

INT_PTR
skkimeConfig_dlgEditServerJisyoOnNotify (
	register HWND		hDlg,
	register UINT		uMsg,
	register WPARAM		wParam,
	register LPARAM		lParam)
{
	register NMHDR	FAR*	pNMHDR	= (NMHDR *) lParam ;
	static TMYMENUITEM		rmi []	= {
		{ MIIM_ID | MIIM_STRING, 0,  TEXT("ǉ"), IDC_BUTTON_ADD_NEW_JISYO_SERVER, },
		{ MIIM_ID | MIIM_STRING, 0,  TEXT("ҏW"), IDC_BUTTON_MODIFY_JISYO_SERVER, },
		{ MIIM_ID | MIIM_STRING, 0,  TEXT("폜"), IDC_BUTTON_REMOVE_JISYO_SERVER, },
	} ;

	switch (pNMHDR->idFrom) {
	case	IDC_LIST_SERVER_DICT:
		if (pNMHDR->code == NM_RCLICK) {
			register NMITEMACTIVATE*	pNM 	= (NMITEMACTIVATE*) lParam ;
			register int	n ;

			n	= skkimeConfig_PopupMenu (hDlg, rmi, ARRAYSIZE (rmi)) ;
			if (n > 0) 
				skkimeConfig_editServerJisyo (hDlg, pNM->iItem, n) ;
		}
		break ;
	default:
		break ;
	}
	return	0 ;
	UNREFERENCED_PARAMETER (uMsg) ;
	UNREFERENCED_PARAMETER (wParam) ;
}

BOOL
skkimeConfig_editServerJisyo (
	register HWND		hDlg,
	register int		nItem,
	register int		nCommand)
{
	register TDlgEditDictArg*	pArg ;
	TDlgEditServerJisyoArg		arg ;
	register TINYMEMMGR*		pMemMgr ;
	register HINSTANCE			hInst ;
	register HWND				hwndLV ;
	MYLVITEM					lvI ;
	register PORTANDHOSTNODE*	pNode ;
	register int				nRetval ;
	MYCHAR						szBuffer [32] ;

	pArg	= (TDlgEditDictArg*) GetWindowLongPtr (hDlg, DWLP_USER) ;
	if (pArg == NULL)
		return	FALSE ;
	pMemMgr	= pArg->m_pMemMgr ;
	hInst	= pArg->m_hInst ;

	hwndLV	= GetDlgItem (hDlg, IDC_LIST_SERVER_DICT) ;
	if (hwndLV == NULL)
		return	FALSE ;
	memset (&lvI, 0, sizeof (lvI)) ;
	lvI.mask		= LVIF_PARAM ;
	lvI.iItem		= nItem ;
	if (! MyListView_GetItem (hwndLV, &lvI)) {
		pNode		= NULL ;
	} else {
		pNode		= (PORTANDHOSTNODE *)lvI.lParam ;
	}
	switch (nCommand) {
	case	IDC_BUTTON_ADD_NEW_JISYO_SERVER:
	case	IDC_BUTTON_MODIFY_JISYO_SERVER:
	{
		register PORTANDHOSTNODE*	pNewNode ;
		register int				nHostname ;
		
		arg.m_strHost [0]	= MYTEXT ('\0') ;
		arg.m_nPort			= 1178 ;
		if (pNode != NULL) {
			if (pNode->m_strHost != NULL)
				Mylstrcpyn (arg.m_strHost, pNode->m_strHost, MAX_PATH) ;
			arg.m_nPort	= pNode->m_nPort ;
		}
		nRetval	= DialogBoxParam (hInst, MAKEINTRESOURCE (IDD_DIALOG_EDIT_JISYOSERVER), hDlg, skkimeConfig_dlgEditJisyoServer, (LPARAM)&arg) ;
		if (nRetval != IDOK)
			return	TRUE ;
		nHostname	= Mylstrlen (arg.m_strHost) ;
		if (arg.m_nPort <= 0 || nHostname <= 0)
			return	FALSE ;
		pNewNode	= skkimeConfig_newPortAndHostNode (pMemMgr, arg.m_strHost, nHostname, arg.m_nPort) ;
		if (pNewNode == NULL) {
			DEBUGPRINTF ((MYTEXT ("out of memory.\n"))) ;
			return	FALSE ;	// out of memory
		}
		skkimeConfig_registerPortAndHostNode   (pNewNode, &pArg->m_lstHost, pNode) ;
		skkimeConfig_unregisterPortAndHostNode (pNode, &pArg->m_lstHost) ;

		memset (&lvI, 0, sizeof (lvI)) ;
		lvI.mask		= LVIF_PARAM | LVIF_STATE ;
		lvI.lParam		= (LPARAM) pNewNode ;
		if (pNode != NULL) {
			lvI.iItem	= nItem ;
			MyListView_SetItem (hwndLV, &lvI) ;
			TinyMalloc_Free (pMemMgr, pNode) ;
		} else {
			nItem		= MyListView_InsertItem (hwndLV, &lvI) ;
		}
		MyListView_SetItemText (hwndLV, nItem, 0, pNewNode->m_strHost) ;
		Mywsprintf (szBuffer, MYTEXT("%d"), pNewNode->m_nPort) ;
		MyListView_SetItemText (hwndLV, nItem, 1, szBuffer) ;
		return	TRUE ;
	}
	case	IDC_BUTTON_REMOVE_JISYO_SERVER:
	{
		if (pNode == NULL)
			return	FALSE ;
		MyListView_DeleteItem (hwndLV, nItem) ;
		skkimeConfig_unregisterPortAndHostNode (pNode, &pArg->m_lstHost) ;
		return	TRUE ;
	}
	default:
		return	FALSE ;
	}
	return	TRUE ;
}

INT_PTR CALLBACK
skkimeConfig_dlgEditJisyoServer (
	register HWND		hDlg,
	register UINT		uMsg,
	register WPARAM		wParam,
	register LPARAM		lParam)
{
	register TDlgEditServerJisyoArg*	pArg ;

	switch (uMsg) {
	case	WM_INITDIALOG:
	{
		register HWND	hwndEditHost, hwndEditPort ;
		MYCHAR			szBuffer [32] ;

		SetWindowLongPtr (hDlg, DWLP_USER, (LONG_PTR) lParam) ;
		pArg			= (TDlgEditServerJisyoArg*) lParam ;
		hwndEditHost	= GetDlgItem (hDlg, IDC_EDIT_HOSTNAME) ;
		hwndEditPort	= GetDlgItem (hDlg, IDC_EDIT_PORTNUM) ;
		if (pArg == NULL || hwndEditHost == NULL || hwndEditPort == NULL)
			break ;
		MySendMessage   (hwndEditHost, EM_LIMITTEXT, (WPARAM) MAX_PATH, 0) ;
		MySetWindowText (hwndEditHost, pArg->m_strHost) ;
		MySendMessage   (hwndEditPort, EM_LIMITTEXT, (WPARAM) 32, 0) ;
		Mywsprintf (szBuffer, MYTEXT ("%d"), pArg->m_nPort) ;
		MySetWindowText (hwndEditPort, szBuffer) ;
		return	TRUE ;
	}
	case	WM_COMMAND:
		switch (LOWORD (wParam)) {
		case	IDOK:
			pArg			= (TDlgEditServerJisyoArg*) GetWindowLongPtr (hDlg, DWLP_USER) ;
			if (pArg != NULL) {
				BOOL	fSuccess ;
				UINT	uVal ;

				MyGetDlgItemText (hDlg, IDC_EDIT_HOSTNAME, pArg->m_strHost, MAX_PATH) ;
				uVal	= GetDlgItemInt  (hDlg, IDC_EDIT_PORTNUM,  &fSuccess, TRUE) ;
				if (fSuccess)
					pArg->m_nPort	= (int)uVal ;
			} else {
				wParam	= IDCANCEL ;
			}
			/*	fall through */
		case	IDCANCEL:
			EndDialog (hDlg, wParam) ;
			return	TRUE ;
		default:
			break ;
		}
		break ;
	default:
		break ;
	}
	return	FALSE ;
}

INT_PTR	CALLBACK
skkimeConfig_dlgEditFileJisyo (
	register HWND		hDlg,
	register UINT		uMsg,
	register WPARAM		wParam,
	register LPARAM		lParam)
{
	register TDlgEditDictArg*	pArg ;
	register HWND	hwndEditPath ;

	switch (uMsg) {
	case	WM_INITDIALOG:
	{
		register UINT	uCheck ;

		SetWindowLongPtr (hDlg, DWLP_USER, (LONG_PTR) lParam) ;
		pArg			= (TDlgEditDictArg*) lParam ;
		hwndEditPath	= GetDlgItem (hDlg, IDC_EDIT_FILE_DICT_PATH) ;
		if (pArg == NULL || hwndEditPath == NULL)
			break ;
		MySendMessage   (hwndEditPath, EM_LIMITTEXT, (WPARAM) MAX_PATH, 0) ;
		MySetWindowText (hwndEditPath, pArg->m_strPath) ;
		uCheck	= (pArg->m_nType == DICTYPE_SORT)? BST_CHECKED : BST_UNCHECKED ;
		CheckDlgButton (hDlg, IDC_CHECK_FILE_DICT_SORTEDP, uCheck) ;
		return	TRUE ;
	}
	case	WM_COMMAND:
	{
		register WORD	wNotifyCode	= HIWORD (wParam) ;
		register WORD	wID			= LOWORD (wParam) ;
		register HWND	hwndCtl		= (HWND) lParam ;

		switch (wID) {
		case	IDC_BUTTON_BROWSE_FILE_DICT_PATH:
		{
			MYOPENFILENAME	ofn ;
			MYCHAR			szPath [MAX_PATH + 1] ;

			pArg	= (TDlgEditDictArg*) GetWindowLongPtr (hDlg, DWLP_USER) ;
			if (pArg == NULL)
				break ;

			if (wNotifyCode == BN_CLICKED) {
				MyGetDlgItemText (hDlg, IDC_EDIT_FILE_DICT_PATH, szPath, MAX_PATH) ;
				szPath [MAX_PATH]	= MYTEXT ('\0') ;
				memset (&ofn, 0, sizeof (ofn)) ;
				ofn.lStructSize			= sizeof (MYOPENFILENAME) ;
				ofn.hwndOwner			= hDlg ;
				ofn.hInstance			= pArg->m_hInst ;
				ofn.lpstrFilter			= MYTEXT ("All\0*.*\0") ;
				ofn.lpstrCustomFilter	= NULL ;
				ofn.nFilterIndex		= 1 ;
				ofn.lpstrFile			= szPath ;
				ofn.nMaxFile			= MAX_PATH ;
				ofn.lpstrFileTitle		= NULL ;
				ofn.lpstrInitialDir		= MYTEXT ("") ;
				ofn.lpstrTitle			= MYTEXT ("t@C̑I") ;
				ofn.Flags				= OFN_HIDEREADONLY | OFN_LONGNAMES | OFN_NONETWORKBUTTON | OFN_PATHMUSTEXIST ;
				ofn.lpstrDefExt			= NULL ;
				ofn.lpfnHook			= NULL ;
				if (MyGetOpenFileName (&ofn) != 0) 
					MySetDlgItemText (hDlg, IDC_EDIT_FILE_DICT_PATH, szPath) ;
			}
			break ;
		}
		case	IDOK:
		{
			pArg			= (TDlgEditDictArg*) GetWindowLongPtr (hDlg, DWLP_USER) ;
			if (pArg != NULL) {
				if (IsDlgButtonChecked (hDlg, IDC_CHECK_FILE_DICT_SORTEDP) == BST_CHECKED) {
					pArg->m_nType	= DICTYPE_SORT ;
				} else {
					pArg->m_nType	= DICTYPE_FILE ;
				}
				MyGetDlgItemText (hDlg, IDC_EDIT_FILE_DICT_PATH, pArg->m_strPath, MAX_PATH) ;
			} else {
				wParam	= IDCANCEL ;
			}
		}
		/*	fall through */
		case	IDCANCEL:
			EndDialog (hDlg, wParam) ;
			return	TRUE ;
		default:
			break ;
		}
		break ;
	}
	default:
		break ;
	}
	return	FALSE ;
}

BOOL
skkimeConfig_setDictionaryItemText (
	register HWND		hwndLV,
	register int		nItem,
	register DICTNODE*	pNode)
{
	static LPMYSTR	rpszDicType []	= {
		MYTEXT ("[U"),		MYTEXT ("T[o"),
		MYTEXT ("t@C"),	MYTEXT ("\[gσt@C"),
	} ;
	MYCHAR			rszBuffer [256] ;
	register LPMYSTR	strArg ;
				
	if (hwndLV == NULL || nItem < 0 || pNode == NULL)
		return	FALSE ;

	switch (pNode->m_nType) {
	case	DICTYPE_SERVER:
		skkimeConfig_hostlist2string (rszBuffer, ARRAYSIZE (rszBuffer), pNode->m_lstHost) ;
		strArg	= rszBuffer ;
		break ;
	case	DICTYPE_FILE:
	case	DICTYPE_SORT:
		strArg	= pNode->m_strPath ;
		break ;
	case	DICTYPE_USER:
		strArg	= MYTEXT ("") ;
		break ;
	default:
		return	FALSE ;
	}
	MyListView_SetItemText (hwndLV, nItem, 0, rpszDicType [pNode->m_nType]) ;
	MyListView_SetItemText (hwndLV, nItem, 1, strArg) ;
	return	TRUE ;
}

PORTANDHOSTNODE*
skkimeConfig_copyPortAndHostList (
	register TINYMEMMGR*		pMemMgr,
	register PORTANDHOSTNODE*	lstHost)
{
	register PORTANDHOSTNODE*	pTopNode ;
	register PORTANDHOSTNODE*	pLastNode ;
	register PORTANDHOSTNODE*	pNode ;
	register PORTANDHOSTNODE*	pOrg ;
	register int				nHostnameLen ;

	pLastNode	= NULL ;
	pOrg		= lstHost ;
	while (pOrg != NULL) {
		nHostnameLen	= Mylstrlen (pOrg->m_strHost) ;
		pNode	= TinyMalloc_Malloc (pMemMgr, sizeof (PORTANDHOSTNODE) + sizeof (MYCHAR) * (nHostnameLen + 1)) ;
		if (pNode == NULL)
			break ;
		pNode->m_strHost	= (LPMYSTR)(pNode + 1) ;
		Mylstrcpy (pNode->m_strHost, pOrg->m_strHost) ;
		pNode->m_nPort		= pOrg->m_nPort ;
		pNode->m_pNext		= NULL ;
		pNode->m_pPrev		= pLastNode ;
		if (pLastNode != NULL)
			pLastNode->m_pNext	= pNode ;
		pLastNode			= pNode ;
		pOrg				= pOrg->m_pNext ;
	}
	if (pLastNode != NULL) {
		register PORTANDHOSTNODE*	pPrevNode ;

		pNode	= pLastNode ;
		while (pPrevNode = pNode->m_pPrev, pPrevNode != NULL)
			pNode	= pPrevNode ;
		pTopNode	= pNode ;
	} else {
		pTopNode	= NULL ;
	}
	return	pTopNode ;
}



BOOL
skkimeConfig_initSearchJisyoList (
	register TINYMEMMGR*		pMemMgr,
	register LPCMYSTR			pstrSubKey,
	register DICTNODE**			plstSearchJisyo,
	register LPMYSTR			strUserJisyoPath)
{
	HKEY				hSubKey ;
	MYCHAR				szRegPathInfo [MAX_PATH + 1] ;
	MYCHAR				rbyData [MAX_REGVALUE_SIZE] ;
	DWORD				dwRegPathInfo, dwType, cbData ;
	register DICTNODE*	pNode ;

	*plstSearchJisyo	= NULL ;

	/*	WXgݒǂށB*/
	if (MyRegOpenKeyEx (HKEY_CURRENT_USER, pstrSubKey, 0, KEY_READ, &hSubKey) == ERROR_SUCCESS) {
		register int	nIndex ;
		int				nDict ;
		register LONG	lResult ;

		for (nIndex = 0 ; nIndex < INT_MAX ; nIndex ++) {
			dwRegPathInfo	= ARRAYSIZE (szRegPathInfo) ;
			cbData			= ARRAYSIZE (rbyData) ;
			lResult	= MyRegEnumValue (hSubKey, (DWORD)nIndex, szRegPathInfo, &dwRegPathInfo, NULL, NULL, NULL, NULL) ;
			if (lResult != ERROR_SUCCESS)
				break ;
			if (skkimeConfig_parseDictNumber (szRegPathInfo, &nDict)) {
				/*	nDict ԍB*/
				lResult	= MyRegQueryValueEx (hSubKey, szRegPathInfo, NULL, &dwType, NULL, &cbData) ;
				if (lResult != ERROR_SUCCESS || dwType != REG_SZ)
					continue ;
				/*	Too big value. */
				if (cbData >= sizeof (rbyData))
					continue ;
				cbData	= sizeof (rbyData) ;
				(void) MyRegQueryValueEx (hSubKey, szRegPathInfo, NULL, NULL, (BYTE *)rbyData, &cbData) ;

				pNode	= skkimeConfig_parseDictionarySetting (pMemMgr, rbyData, nDict) ;
				if (pNode != NULL) 
					skkimeConfig_insertDictionaryInSearchList (plstSearchJisyo, pNode) ;
			}
		}
		lResult	= MyRegQueryValueEx (hSubKey, REGKEY_USERJISYO, NULL, &dwType, NULL, &cbData) ;
		if (lResult != ERROR_SUCCESS ||
			(dwType != REG_SZ && dwType != REG_EXPAND_SZ) ||
			cbData > sizeof (MYCHAR) * MAX_PATH) {
			Mylstrcpy (strUserJisyoPath, SKKIME_DEFAULT_USERJISYOPATH) ;
		} else {
			cbData	= MAX_PATH ;
			MyRegQueryValueEx (hSubKey, REGKEY_USERJISYO, NULL, &dwType, (BYTE *)strUserJisyoPath, &cbData) ;
			strUserJisyoPath [MAX_PATH]	= MYTEXT ('\0') ;
		}
		MyRegCloseKey (hSubKey) ;
		return	TRUE ;
	}

	/*	ftHg̐ݒB
	 *	ftHg̐ݒł́A[Ǔ擪ɂ邾H
	 */
	pNode	= skkimeConfig_parseDictionarySetting (pMemMgr, SKKIME_DEFAULT_JISYO_SETTING, 0) ;
	if (pNode == NULL)
		return	FALSE ;
	skkimeConfig_insertDictionaryInSearchList (plstSearchJisyo, pNode) ;

	/*	[ŨpX̓ftHǵA
	 *		%USERPROFILE%\Application Data\skk\skki1_0\skki1_0u.dic
	 *	łB
	 */
	Mylstrcpy (strUserJisyoPath, SKKIME_DEFAULT_USERJISYOPATH) ;
	return	TRUE ;
}

BOOL
skkimeConfig_updateSearchJisyoList (
	register LPCMYSTR		pstrSubKey,
	register DICTNODE*		lstSearchJisyo,
	register LPMYSTR		strUserJisyoPath)
{
	MYCHAR				rbufTemp [MAX_PATH + 1] ;
	MYCHAR				szData [MAX_REGVALUE_SIZE] ;
	HKEY				hSubKey ;
	register BOOL		fRetval	= TRUE ;
	register DICTNODE*	pNode ;
	register int		nLength, nDest, nDict ;
	register LPCMYSTR	strFormat ;
	register LPMYSTR	pDest ;
	register PORTANDHOSTNODE*	pHostPortNode ;

	if (! skkimeConfig_CreateKey (pstrSubKey, TRUE, &hSubKey)) {
		DEBUGPRINTF ((MYTEXT ("skkimeConfig_updateSearchJisyoList: create-key failed(\"%s\")\n"), pstrSubKey)) ;
		return	FALSE ;
	}

	pNode	= lstSearchJisyo ;
	nDict	= 0 ;
	while (pNode != NULL) {
		pDest	= szData ;
		nDest	= ARRAYSIZE (szData) ;
		switch (pNode->m_nType) {
		case	DICTYPE_USER:
			Mylstrcpy (pDest, MYTEXT ("user=")) ;
			pDest	+= 6 ;	/* lstrlen ("user=") == 6 */
			nDest	-= 6 ;
			break ;
		case	DICTYPE_SERVER:
			Mylstrcpy (pDest, MYTEXT ("server=")) ;
			pDest	+= 7 ;	/* lstrlen ("server=") == 7 */
			nDest	-= 7 ;
			pHostPortNode	= pNode->m_lstHost ;
			strFormat		= MYTEXT ("%d/%s") ;
			while (pHostPortNode != NULL) {
				if (pHostPortNode->m_nPort >= 0 &&
					pHostPortNode->m_strHost != NULL &&
					Mylstrlen (pHostPortNode->m_strHost) < (MAX_PATH - 16)) {
					Mywsprintf (rbufTemp, strFormat, pHostPortNode->m_nPort, pHostPortNode->m_strHost) ;
					nLength	= Mylstrlen (rbufTemp) ;
					if (nDest > nLength) {
						Mylstrcpy (pDest, rbufTemp) ;
						pDest		+= nLength ;
						nDest		-= nLength ;
						strFormat	= MYTEXT (",%d/%s") ;
					} else {
						DEBUGPRINTF ((MYTEXT ("Too many (port/host)s\n"))) ;
						/* ȏۑłȂƎvƏ𔲂B*/
						break ;
					}
				}
				pHostPortNode	= pHostPortNode->m_pNext ;
			}
			/*	⏕̐ݒ͂邩H */
			if (pNode->m_strPath != NULL && pNode->m_strPath [0] != MYTEXT ('\0')) {
				nLength	= Mylstrlen (pNode->m_strPath) ;
				if (nDest > (nLength + 8)) {	/* lstrlen (",aux:??=") == 8 */
					Mylstrcpy (pDest, MYTEXT (",aux:")) ;
					pDest	+= 5 ;
					*pDest	++ = (MYCHAR)((pNode->m_fSortedAuxJisyo)? MYTEXT ('s') : MYTEXT ('f')) ;
					*pDest	++ = (MYCHAR)((pNode->m_fEnableAuxJisyo)? MYTEXT ('e') : MYTEXT ('d')) ;
					*pDest	++ = MYTEXT ('=') ;
					nDest	-= 8 ;
					Mylstrcpy (pDest, pNode->m_strPath) ;
					pDest	+= nLength ;
					nDest	-= nLength ;
				}
			}
			break ;
		case	DICTYPE_SORT:
			Mylstrcpy (pDest, MYTEXT ("sort=")) ;
			goto	dictype_file_common ;
		case	DICTYPE_FILE:
			Mylstrcpy (pDest, MYTEXT ("file=")) ;
		dictype_file_common:
			pDest	+= 5 ;	/* lstrlen ("sort=") == 5 && lstrlen ("file=") == 5 */
			nDest	-= 5 ;
			nLength	= Mylstrlen (pNode->m_strPath) ;
			if ((nLength + (pDest - szData) + 1) >= MAX_REGVALUE_SIZE) {
				DEBUGPRINTF ((MYTEXT ("Too long dictionary path (%s).\n"), pNode->m_strPath)) ;
				goto	skip ;
			}
			Mylstrcpy (pDest, pNode->m_strPath) ;
			pDest	+= nLength ;
			nDest	-= nLength ;
			break ;
		default:
			DEBUGPRINTF ((MYTEXT ("Invalid dictionary type.\n"))) ;
			goto	skip ;
		}
		Mywsprintf (rbufTemp, MYTEXT ("Dict%d"), nDict) ;
		if (MyRegSetValueEx (hSubKey, rbufTemp, 0, REG_SZ, (BYTE *)szData, sizeof (MYCHAR) * ((pDest - szData) + 1)) != ERROR_SUCCESS) {
			DEBUGPRINTF ((MYTEXT ("RegSetValueEx: failed (\"%s\", \"%s\")\n"), rbufTemp, szData)) ;
			fRetval	= FALSE ;
		} else {
			DEBUGPRINTF ((MYTEXT ("RegSetValueEx: \"%s\" <= \"%s\")\n"), rbufTemp, szData)) ;
			fRetval	= TRUE ;
		}
		nDict	++ ;
	skip:
		pNode	= pNode->m_pNext ;
	}

	nLength	= Mylstrlen (strUserJisyoPath) ;
	if (nLength > 0) {
		if (MyRegSetValueEx (hSubKey, REGKEY_USERJISYO, 0, REG_SZ, (BYTE *)strUserJisyoPath, sizeof (MYCHAR) * (nLength + 1)) != ERROR_SUCCESS) {
			DEBUGPRINTF ((MYTEXT ("RegSetValueEx: failed (\"%s\")\n"), strUserJisyoPath)) ;
			fRetval	= FALSE ;
		}
	}
	MyRegCloseKey (hSubKey) ;
	return	fRetval ;
}

BOOL
skkimeConfig_parseDictNumber (
	register LPCMYSTR		pstrString,
	register int*			pnDict)
{
	register int	nValue ;

	if (Mylstrcmpn (pstrString, MYTEXT ("Dict"), 4))
		return	FALSE ;

	pstrString	+= 4 ;
	if (*pstrString == MYTEXT ('\0'))
		return	FALSE ;

	nValue	= 0 ;
	while (*pstrString != MYTEXT ('\0')) {
		if (! (MYTEXT ('0') <= *pstrString && *pstrString <= MYTEXT ('9')))
			return	FALSE ;
		nValue	= nValue * 10 + (*pstrString - MYTEXT ('0')) ;
		if (nValue < 0)
			return	FALSE ;
		pstrString	++ ;
	}
	*pnDict	= nValue ;
	return	TRUE ;
}

/*	̐ݒ蕶͎̒ʂłF
 *		T[ȍꍇAserver=port1/host1,port2/host2,aux=aux-dictionarypath
 *		(͍Ō̗vf)
 *		[ȔꍇAuser=(͂Ȃ)
 *		̑t@CAfile=path
 *
 *	aux:??= ƉρBaux:s  sort ĂBaux:f ł fileBaux:e  enable ŁA
 *	aux:d  disableBgݍ킹ł́A
 *		(s or f) and (e or d) ɂȂB
 *	aux:se Ƃ aux:fe Ƃ aux:sd Ƃ aux:fd ƂB
 */
DICTNODE*
skkimeConfig_parseDictionarySetting (
	register TINYMEMMGR*	pMemMgr,
	register LPCMYSTR		pstrSetting,
	register int			nIndex)
{
	DICTNODE*	pNode	= NULL ;

	if (! Mylstrcmpn (pstrSetting, MYTEXT ("server="), 7)) {
		register LPCMYSTR	ptr	;
		register LPCMYSTR	pStart ;
		register LPCMYSTR	pAuxPath	= NULL ;
		register int		nServer ;
		register PORTANDHOSTNODE*	pHostPortNode ;
		register PORTANDHOSTNODE*	plstHostPort	= NULL ;
		register BOOL		fEnableAuxJisyo	= FALSE ;
		register BOOL		fSortedAuxJisyo	= FALSE ;

		nServer	= 0 ;
		pStart	= pstrSetting + 7 ;
		while (*pStart != MYTEXT ('\0')) {
			ptr		= pStart ;
			while (*ptr != MYTEXT ('\0') && *ptr != MYTEXT (',')) 
				ptr	++ ;
			/*	⏕̐ݒ肪tꍇAɑT[o
			 *	ݒ݂͑ȂB*/
			if (! Mylstrcmpn (pStart, MYTEXT ("aux:"), 4)) {
				pStart		+= 4 ;
				if (*pStart == MYTEXT('s') || *pStart == MYTEXT('f')) {
					fSortedAuxJisyo	= (*pStart == MYTEXT('s'))? TRUE : FALSE ;
					pStart	++ ;
				}
				if (*pStart == MYTEXT('e') || *pStart == MYTEXT('d')) {
					fEnableAuxJisyo	= (*pStart == MYTEXT ('e'))? TRUE : FALSE ;
					pStart	++ ;
				}
				/*	̏ꍇɂ̓G[B̓VrAB*/
				if (*pStart == MYTEXT('=')) {
					pAuxPath	= pStart + 1 ;
				} else {
					pAuxPath	= NULL ;
				}
				break ;
			}
			/*	port/host ̃T[o̐ݒ𔲂oB*/
			pHostPortNode	= skkimeConfig_parsePortAndHost (pMemMgr, pStart, ptr - pStart) ;
			if (pHostPortNode != NULL) {
				pHostPortNode->m_pNext	= plstHostPort ;
				if (plstHostPort != NULL)
					plstHostPort->m_pPrev	= pHostPortNode ;
				plstHostPort			= pHostPortNode ;
			}
			if (*ptr == MYTEXT ('\0'))
				break ;
			pStart	= ptr + 1 ;
		}
		pNode	= skkimeConfig_newDictionaryNode (pMemMgr, DICTYPE_SERVER, pAuxPath) ;
		pNode->m_fEnableAuxJisyo	= fEnableAuxJisyo ;
		pNode->m_fSortedAuxJisyo	= fSortedAuxJisyo ;
		pNode->m_lstHost			= plstHostPort ;
	} else if (! Mylstrcmpn (pstrSetting, MYTEXT ("user="), 5)) {
		pNode	= skkimeConfig_newDictionaryNode (pMemMgr, DICTYPE_USER, NULL) ;
	} else if (! Mylstrcmpn (pstrSetting, MYTEXT ("file="), 5) ||
			   ! Mylstrcmpn (pstrSetting, MYTEXT ("sort="), 5)) {
		register int	nType ;

		nType	= Mylstrcmpn (pstrSetting, MYTEXT ("sort="), 5) ? DICTYPE_FILE : DICTYPE_SORT ;
		pstrSetting	+= 5 ;
		if (*pstrSetting == MYTEXT ('\0'))
			return	NULL ;
		pNode	= skkimeConfig_newDictionaryNode (pMemMgr, nType, pstrSetting) ;
	}
	if (pNode != NULL) {
		pNode->m_nIndex		= nIndex ;
		pNode->m_pNext		= NULL ;
		pNode->m_pPrev		= NULL ;
	}
	return	pNode ;
}	

PORTANDHOSTNODE*
skkimeConfig_parsePortAndHost (
	register TINYMEMMGR*	pMemMgr,
	register LPCMYSTR		pstrPortAndHost,
	register int			nPortAndHost)
{
	register LPCMYSTR	ptr		= pstrPortAndHost ;
	register int		nPort	= 0 ;
	register LPCMYSTR	pHostname ;
	register int		nHostnameLen ;

	if (*ptr != MYTEXT ('/')) {
		while (nPortAndHost > 0 && *ptr != MYTEXT ('/')) {
			if (! (MYTEXT ('0') <= *ptr && *ptr <= MYTEXT ('9')))
				return	NULL ;
			nPort	= nPort * 10 + (*ptr - MYTEXT ('0')) ;
			ptr				++ ;
			nPortAndHost	-- ;
		}
		if (nPort < 0 || nPort >= 65536)
			return	NULL ;
	} else {
		nPort	= SKKSERV_DEFAULT_PORTNUM ;
	}
	if (nPortAndHost <= 0 || *ptr != MYTEXT ('/'))
		return	NULL ;

	/*	"/" ǂݔ΂B*/
	nPortAndHost	-- ;
	ptr				++ ;

	nHostnameLen	= nPortAndHost ;
	pHostname		= ptr ;
	if (nHostnameLen <= 0)
		return	NULL ;

	/*	ȉASāAƌƂ낾Ahostname ƂĕsK؂ȕ
	 *	Ύ̂ĂĂ炤B^ʖڂȃp[Y͂ȂB
	 *	host part ɗp\ȕ HTTP-URL  RFC QlɂĂB*/
	while (nPortAndHost > 0 &&
		   ((MYTEXT ('0') <= *ptr && *ptr <= MYTEXT ('9')) ||
			(MYTEXT ('a') <= *ptr && *ptr <= MYTEXT ('z')) ||
			(MYTEXT ('A') <= *ptr && *ptr <= MYTEXT ('Z')) ||
			*ptr == MYTEXT ('-') ||
			*ptr == MYTEXT ('.'))) {
		ptr				++ ;
		nPortAndHost	-- ;
	}
	if (nPortAndHost != 0)
		return	NULL ;
	
	return	skkimeConfig_newPortAndHostNode (pMemMgr, pHostname, nHostnameLen, nPort) ;
}

BOOL
skkimeConfig_insertDictionaryInSearchList (
	register DICTNODE**	plstTop,
	register DICTNODE*	pNewNode)
{
	register DICTNODE*	pNode ;
	register DICTNODE*	pPrevNode ;

	pNode		= *plstTop ;
	pPrevNode	= NULL ;
	while (pNode != NULL && pNode->m_nIndex < pNewNode->m_nIndex) {
		pPrevNode	= pNode ;
		pNode		= pNode->m_pNext ;
	}
	if (pPrevNode != NULL) {
		pPrevNode->m_pNext	= pNewNode ;
	} else {
		*plstTop			= pNewNode ;
	}
	pNewNode->m_pPrev	= pPrevNode ;
	pNewNode->m_pNext	= pNode ;
	if (pNode != NULL) 
		pNode->m_pPrev	= pNewNode ;
	return	TRUE ;
}

DICTNODE*
skkimeConfig_newDictionaryNode (
	register TINYMEMMGR*	pMemMgr,
	register int			nType,
	register LPCMYSTR		strPath)
{
	register DICTNODE*	pNode ;
	register int		nLength ;

	/*	MAX_PATH ȏ PATH ݂̑͋ȂB*/
	nLength	= (strPath != NULL)? (Mylstrlen (strPath) + 1) : 0 ;
	if (nLength > MAX_PATH)
		return	NULL ;
	pNode	= TinyMalloc_Malloc (pMemMgr, sizeof (DICTNODE) + sizeof (MYCHAR) * nLength) ;
	if (strPath != NULL) {
		pNode->m_strPath	= (LPMYSTR)(pNode + 1) ;
		Mylstrcpy (pNode->m_strPath, strPath) ;
	} else {
		pNode->m_strPath	= NULL ;
	}
	pNode->m_nIndex		= 0 ;
	pNode->m_pPrev		= NULL ;
	pNode->m_pNext		= NULL ;
	pNode->m_lstHost	= NULL ;
	pNode->m_nType		= nType ;
	return	pNode ;
}

BOOL
skkimeConfig_registerDictionaryNode (
	register DICTNODE**		plstTop,
	register DICTNODE*		pNode,
	register DICTNODE*		pPrevNode)
{
	register DICTNODE*	pNextNode ;

	if (pNode == NULL || plstTop == NULL)
		return	FALSE ;

	if (pPrevNode == NULL) {
		pNextNode				= *plstTop ;
		pNode->m_pNext			= pNextNode ;
		*plstTop				= pNode ;
		pNode->m_pPrev			= NULL ;
	} else {
		pNextNode				= pPrevNode->m_pNext ;
		pPrevNode->m_pNext		= pNode ;
		pNode->m_pNext			= pNextNode ;
		pNode->m_pPrev			= pPrevNode ;
	}
	if (pNextNode != NULL)
		pNextNode->m_pPrev		= pNode ;
	return	TRUE ;
}

BOOL
skkimeConfig_unregisterDictionaryNode (
	register DICTNODE**		plstTop,
	register DICTNODE*		pNode)
{
	register DICTNODE*	pPrevNode ;
	register DICTNODE*	pNextNode ;

	if (pNode == NULL || plstTop == NULL)
		return	FALSE ;
	pPrevNode	= pNode->m_pPrev ;
	pNextNode	= pNode->m_pNext ;
	if (pPrevNode != NULL) {
		pPrevNode->m_pNext		= pNextNode ;
	} else {
		// pNode == slstSkkinputSearchJisyo
		*plstTop				= pNextNode ;
	}
	if (pNextNode != NULL) {
		pNextNode->m_pPrev		= pPrevNode ;
	}
	return	TRUE ;
}

void
skkimeConfig_deleteDictionaryNode (
	register TINYMEMMGR*	pMemMgr,
	register DICTNODE*		pNode)
{
	if (pNode == NULL)
		return ;
	if (pNode->m_lstHost != NULL) {
		skkimeConfig_deleteHostList (pMemMgr, pNode->m_lstHost) ;
		pNode->m_lstHost	= NULL ;
	}
	TinyMalloc_Free (pMemMgr, pNode) ;
	return ;
}

void
skkimeConfig_deleteHostList (
	register TINYMEMMGR*		pMemMgr,
	register PORTANDHOSTNODE*	lstHost)
{
	register PORTANDHOSTNODE*	pNode ;
	register PORTANDHOSTNODE*	pNextNode ;

	pNode	= lstHost ;
	while (pNode != NULL) {
		pNextNode	= pNode->m_pNext ;
		TinyMalloc_Free (pMemMgr, pNode) ;
		pNode		= pNextNode ;
	}
	return ;
}

PORTANDHOSTNODE*
skkimeConfig_newPortAndHostNode (
	register TINYMEMMGR*	pMemMgr,
	register LPCMYSTR		strHost,
	register int			nHost,
	register int			nPort)
{
	register PORTANDHOSTNODE*	pNode ;

	pNode	= TinyMalloc_Malloc (pMemMgr, sizeof (PORTANDHOSTNODE) + sizeof (MYCHAR) * (nHost + 1)) ;
	pNode->m_pPrev		= NULL ;
	pNode->m_pNext		= NULL ;
	pNode->m_strHost	= (LPMYSTR)(pNode + 1) ;
	pNode->m_nPort		= nPort ;
	memcpy (pNode->m_strHost, strHost, sizeof (MYCHAR) * nHost) ;
	*(pNode->m_strHost + nHost)	= TEXT ('\0') ;
	return	pNode ;
}

BOOL
skkimeConfig_registerPortAndHostNode (
	register PORTANDHOSTNODE*	pNode,
	register PORTANDHOSTNODE**	plstPortAndHost,
	register PORTANDHOSTNODE*	pPrevNode)
{
	register PORTANDHOSTNODE*	pNextNode ;

	if (pNode == NULL || plstPortAndHost == NULL)
		return	FALSE ;
	if (pPrevNode == NULL) {
		pNextNode			= *plstPortAndHost ;
		*plstPortAndHost	= pNode ;
	} else {
		pNextNode			= pPrevNode->m_pNext ;
		pPrevNode->m_pNext	= pNode ;
	}
	if (pNextNode != NULL) {
		pNextNode->m_pPrev	= pNode ;
	}
	pNode->m_pPrev			= pPrevNode ;
	pNode->m_pNext			= pNextNode ;
	return	TRUE ;
}

BOOL
skkimeConfig_unregisterPortAndHostNode (
	register PORTANDHOSTNODE*	pNode,
	register PORTANDHOSTNODE**	plstPortAndHost)
{
	register PORTANDHOSTNODE*	pPrevNode ;
	register PORTANDHOSTNODE*	pNextNode ;

	if (pNode == NULL || plstPortAndHost == NULL)
		return	FALSE ;
	pPrevNode	= pNode->m_pPrev ;
	pNextNode	= pNode->m_pNext ;
	if (pPrevNode == NULL) {
		*plstPortAndHost	= pNextNode ;
	} else {
		pPrevNode->m_pNext	= pNextNode ;
	}
	if (pNextNode != NULL) {
		pNextNode->m_pPrev	= pPrevNode ;
	}
	return	TRUE ;
}

BOOL
skkimeConfig_hostlist2string (
	register LPMYSTR			pDest,
	register int				nDest,
	register PORTANDHOSTNODE*	plstHost)
{
	register PORTANDHOSTNODE*	pNode ;
	register LPCMYSTR	pszFmt ;
	MYCHAR				szBuffer [256] ;
	register int		nLength ;

	pNode	= plstHost ;
	pszFmt	= MYTEXT ("%s/%d") ;
	while (pNode != NULL) {
		/*	32bit integer  10 iŕ\10ΑvH
		 *	16bit -> 65536 < 70000, 70000^2 = 4900000000 -> 10B}CiX
		 *	\ 11B
		 *	3  " ,/" 3B
		 */
		if (Mylstrlen (pNode->m_strHost) < (256 - 11 - 3) &&
			pNode->m_nPort >= 0) {
			Mywsprintf (szBuffer, pszFmt, pNode->m_strHost, pNode->m_nPort) ;
			nLength	= Mylstrlen (szBuffer) ;
			if (nDest < nLength) {
				memcpy (pDest, szBuffer, sizeof (MYCHAR) * nLength) ;
				pDest	+= nLength ;
				nDest	-= nLength ;
				*pDest	= MYTEXT ('\0') ;
				pszFmt	= MYTEXT (", %s/%d") ;
			} else {
				memcpy (pDest, szBuffer, sizeof (MYCHAR) * (nDest - 1)) ;
				*(pDest + nDest - 1)	= MYTEXT ('\0') ;
				break ;
			}
		}
		pNode	= pNode->m_pNext ;
	}
	return	TRUE ;
}

BOOL
skkimeConfig_dlgDictionaryMoveSelectedNode (
	register HWND				hDlg,
	register SkkimeConfigArg*	pConfArg,
	register BOOL				fPrev)
{
	register HWND			hwndLV ;
	register DICTNODE**		plstJisyo ;
	register DICTNODE*		pMoveNode ;
	MYLVITEM				lvI ;
	register int			nItem ;
	MYLVFINDINFO			lvfI ;

	hwndLV			= GetDlgItem (hDlg, IDC_LIST_SEARCHDICT) ;
	if (hwndLV == NULL)
		return	FALSE ;
	
	memset (&lvI, 0, sizeof (lvI)) ;
	lvI.mask		= LVIF_PARAM ;
	nItem			= MyListView_GetSelectionMark (hwndLV) ;
	lvI.iItem		= nItem ;
	if (! MyListView_GetItem (hwndLV, &lvI)) {
		DebugPrintf (TEXT ("MyListView_GetItem failed (item:%d)\n"), nItem) ;
		return	FALSE ;
	}

	pMoveNode	= (DICTNODE*) lvI.lParam ;
	if (pMoveNode == NULL) {
		return	FALSE ;
	}

	memset (&lvfI, 0, sizeof (lvfI)) ;
	lvfI.flags		= LVFI_PARAM ;
	if (fPrev) {
		if (pMoveNode->m_pPrev != NULL) {
			register DICTNODE*	pPrevNode	= pMoveNode->m_pPrev ;
			register int		nPrevItem ;
			
			lvfI.lParam		= (LPARAM) pPrevNode ;
			nPrevItem		= MyListView_FindItem (hwndLV, -1, &lvfI) ;
			if (nPrevItem >= 0) {
				plstJisyo	= &pConfArg->m_lstSearchJisyo ;
				skkimeConfig_unregisterDictionaryNode (plstJisyo, pMoveNode) ;
				skkimeConfig_registerDictionaryNode   (plstJisyo, pMoveNode, pPrevNode->m_pPrev) ;
				
				MyListView_DeleteItem (hwndLV, nItem) ;
				memset (&lvI, 0, sizeof (lvI)) ;
				lvI.mask		= LVIF_PARAM ;
				lvI.iItem		= nPrevItem ;
				lvI.lParam		= (LPARAM) pMoveNode ;
				nItem			= MyListView_InsertItem (hwndLV, &lvI) ;
				skkimeConfig_setDictionaryItemText (hwndLV, nItem, pMoveNode) ;
				MyListView_SetSelectionMark (hwndLV, nItem) ;
				MyListView_SetItemState (hwndLV, nItem, LVIS_SELECTED, LVIS_SELECTED) ;
			}
		}
	} else {
		if (pMoveNode->m_pNext != NULL) {
			register DICTNODE*	pNextNode	= pMoveNode->m_pNext ;
			register int		nNextItem ;

			lvfI.lParam		= (LPARAM) pNextNode ;
			nNextItem		= MyListView_FindItem (hwndLV, -1, &lvfI) ;
			if (nNextItem >= 0) {
				plstJisyo	= &pConfArg->m_lstSearchJisyo ;
				skkimeConfig_unregisterDictionaryNode (plstJisyo, pMoveNode) ;
				skkimeConfig_registerDictionaryNode   (plstJisyo, pMoveNode, pNextNode) ;
				
				MyListView_DeleteItem (hwndLV, nItem) ;
				memset (&lvI, 0, sizeof (lvI)) ;
				lvI.mask		= LVIF_PARAM ;
				lvI.iItem		= nNextItem ;
				lvI.lParam		= (LPARAM) pMoveNode ;
				nItem			= MyListView_InsertItem (hwndLV, &lvI) ;
				skkimeConfig_setDictionaryItemText (hwndLV, nItem, pMoveNode) ;
				MyListView_SetSelectionMark (hwndLV, nItem) ;
				MyListView_SetItemState (hwndLV, nItem, LVIS_SELECTED, LVIS_SELECTED) ;
			}
		}
	}
	return	TRUE ;
}

